#' A subclass of LoadComponentDataSource that represents a database connection
#'
#' @description
#' A subclass must implement `private$get_full_info()`. A subclass can optionally
#' set `private$.uses_schema = FALSE` if it represents a database that does not
#' use schemas, and can optionally set `private$.limit_type` to "top" or "fetch"
#' if it represents a database that doesn't support the `LIMIT` syntax.
#'
#' @field database The database selected
#' @field schema The schema selected
#' @field table The table selected
#' @field uses_schema Whether this database type uses the concept of a schema
#' @field fixed_database If the config provides a database, the user cannot change it
#' @field fixed_schema If the config provides a schema, the user cannot change it
#' @field all_table_info Dataframe with 4 columns: database, schema, table, rows
#' @field limit_type The type of row limiting syntax supported in this database
DatabaseDataSource <- R6::R6Class(
  "DatabaseDataSource",
  inherit = LoadComponentDataSource,

  private = list(
    .database = NULL,
    .schema = NULL,
    .table = NULL,

    .uses_schema = TRUE,

    .all_table_info = NULL,

    .limit_type = "limit",

    get_full_info = function() {
      stop("private$get_full_info() must be implemented by the subclass and return a dataframe with 4 columns: database, schema, table, rows")
    },

    get_name = function() {
      make.names(self$table)
    },

    get_code = function() {
      nrows <- subset(self$get_tables(), table == self$table, select = "rows", drop = TRUE)
      limit_pre <- ""
      limit_post <- ""
      if (length(nrows) != 0 && nrows[1] > DATASET_MAX_ROWS_TO_LOAD) {
        if (self$limit_type == "limit") {
          limit_post <- paste(" LIMIT", DATASET_MAX_ROWS_TO_LOAD)
        } else if (self$limit_type == "top") {
          limit_pre <- paste("TOP", DATASET_MAX_ROWS_TO_LOAD, "")
        } else if (self$limit_type == "fetch") {
          limit_post <- paste(" FETCH FIRST", DATASET_MAX_ROWS_TO_LOAD, "ROWS ONLY")
        }
      }

      if (self$fixed_database) {
        db <- ""
      } else {
        db <- paste0(self$database, ".")
      }
      if (!self$uses_schema || self$fixed_schema) {
        schema <- ""
      } else {
        schema <- paste0(self$schema, ".")
      }

      code <- glue::glue(
        'client <- { ddr_function("datasource_client", text = TRUE) }()',
        'sql <- "SELECT {limit_pre}* FROM { db }{ schema }{ self$table }{ limit_post }"',
        'result <- { ddr_function("query", text = TRUE) }(client, { shQuote(self$dds$name, type = "cmd") }, sql)',
        'rm(client)',
        'as.data.frame(result)',
        .sep = '\n'
      )
      code
    },

    get_data = function() {
      suppressWarnings(eval(parse(text = private$.code)))
    }
  ),

  active = list(
    database = function() {
      private$.database
    },
    schema = function() {
      private$.schema
    },
    table = function() {
      private$.table
    },

    uses_schema = function() {
      private$.uses_schema
    },

    fixed_database = function() {
      !is.null(self$dds$config$database)
    },
    fixed_schema = function() {
      !is.null(self$dds$config$schema)
    },

    all_table_info = function() {
      if (is.null(private$.all_table_info)) {
        info <- private$get_full_info()
        private$.all_table_info <- info
      }
      private$.all_table_info
    },

    limit_type = function() {
      private$.limit_type
    }
  ),

  public = list(

    #' @description
    #' Create a new instance
    #'
    #' @param dds The DominoDataSource associated with this database
    initialize = function(dds) {
      super$initialize(dds = dds)
      if (!is.null(dds$config$database)) {
        private$.database <- dds$config$database
      }
      if (!is.null(dds$config$schema)) {
        private$.schema <- dds$config$schema
      }
      invisible(self)
    },

    #' @description
    #' Print a string representation
    print = function() {
      super$print()

      if (!self$fixed_database && !is.null(self$database)) {
        cat("    Database:", self$database)
      }
      if (!self$fixed_schema && !is.null(self$schema)) {
        cat("    Schema:", self$schema)
      }
      if (!is.null(self$table)) {
        cat("    Table:", self$table)
      }
      cat("\n")
    },

    #' @description
    #' Get a list of databases available in this datasource
    get_databases = function() {
      info <- self$all_table_info
      unique(info$database)
    },

    #' @description
    #' Set the database
    #'
    #' @param x A database name
    set_database = function(x) {
      if (is.null(x) || x == "") x <- NULL
      private$.database <- x
      private$.invalidate()
      invisible(self)
    },

    #' @description
    #' Get a list of schemas available for the given database
    get_schemas = function() {
      if (is.null(self$database)) {
        return(list())
      }
      info <- self$all_table_info
      filtered_info <- info[info$database == self$database, ]
      unique(filtered_info$schema)
    },

    #' @description
    #' Set the schema
    #'
    #' @param x A schema name
    set_schema = function(x) {
      if (is.null(x) || x == "") x <- NULL
      private$.schema <- x
      private$.invalidate()
      invisible(self)
    },

    #' @description
    #' Get a list of tables available for the given database/schema
    get_tables = function() {
      if (is.null(self$database) || (self$uses_schema && is.null(self$schema))) {
        return(data.frame(table = character(0), rows = numeric(0)))
      }
      info <- self$all_table_info
      filtered_info <- info[info$database == self$database, ]
      if (self$uses_schema) {
        filtered_info <- filtered_info[filtered_info$schema == self$schema, ]
      }
      filtered_info[, c("table", "rows")]
    },

    #' @description
    #' Set the table
    #'
    #' @param x A table name
    set_table = function(x) {
      if (is.null(x) || x == "") x <- NULL
      private$.table <- x
      private$.invalidate()
      if (is.null(x)) {
        private$.name <- private$.code <- private$.data <- private$.error <- NULL
      } else {
        private$run()
      }
      invisible(self)
    }
  )
)

#' Shiny module: Database data source
#'
#' @param id Unique ID for the module
#' @param ds A DatabaseDataSource object
#' @return List of reactives:
#'   - name
#'   - data
#'   - code
#'   - error
#' @name database_datasource
NULL

#' @rdname database_datasource
database_datasource_ui <- function(id) {
  ns <- NS(id)
  tagList(
    shinyjs::useShinyjs(),
    uiOutput(ns("error")),
    shinyjs::hidden(
      div(
        id = ns("database-params"),
        fluidRow(
          column(
            6,
            shinyWidgets::pickerInput(ns("database"), "Database", choices = NULL, multiple = TRUE, options = list(`max-options` = 1), width = "100%")
          ),
          column(
            6,
            shinyWidgets::pickerInput(ns("schema"), "Schema", choices = NULL, multiple = TRUE, options = list(`max-options` = 1), width = "100%")
          )
        ),
        fluidRow(
          column(
            12,
            shinyWidgets::pickerInput(ns("table"), "Table", choices = NULL, multiple = TRUE, options = list(`max-options` = 1), width = "100%")
          ),
          column(
            12,
            shinyjs::hidden(
              shinyWidgets::alert(
                id = ns("too-many-rows-error"),
                status = "warning",
                paste("This table has too many rows. Only", prettyNum(DATASET_MAX_ROWS_TO_LOAD, big.mark = ","), "rows will be loaded.")
              )
            )
          )
        )
      )
    )
  )
}

#' @rdname database_datasource
database_datasource_server <- function(id, ds) {
  if (!inherits(ds, DatabaseDataSource$classname)) {
    stop("database_datasource_server: Only <DatabaseDataSource> objects are supported")
  }

  moduleServer(
    id,
    function(input, output, session) {

      info <- shinycssloaders::showPageSpinner(
        try(ds$all_table_info, silent = TRUE),
        caption = span("Getting info from", tags$em(ds$dds$name))
      )

      if (inherits(info, "try-error")) {
        error <- attr(info, "condition")$message
        if (grepl("KeyError", error)) {
          error <- paste(error, "(This may be due to an incompatibility between your Domino version and the python dominodatalab-data module version)")
        }
        output$error <- renderUI({
          div(class = "alert alert-danger", icon("exclamation-sign", lib = "glyphicon"), error)
        })
        return()
      }

      shinyjs::show("database-params")
      uses_schema <- ds$uses_schema
      if (!uses_schema) {
        shinyjs::hide("schema")
      }

      shinyWidgets::updatePickerInput(session, "database", choices = ds$get_databases())

      if (ds$fixed_database) {
        shinyWidgets::updatePickerInput(session, "database", choices = ds$database, selected = ds$database)
        shinyjs::disable("database")
      }
      if (ds$fixed_schema) {
        shinyWidgets::updatePickerInput(session, "schema", choices = ds$schema, selected = ds$schema)
        shinyjs::disable("schema")
      }

      ds <- ds$reactive()

      observeEvent(input$database, ignoreNULL = FALSE, {
        ds()$set_database(input$database)
        if (uses_schema) {
          if (!ds()$fixed_schema) {
            schemas <- ds()$get_schemas() %||% list()
            shinyWidgets::updatePickerInput(session, "schema", choices = schemas)
          }
        } else {
          tables <- ds()$get_tables()$table %||% list()
          shinyWidgets::updatePickerInput(session, "table", choices = tables)
        }
      })

      observeEvent(input$schema, ignoreNULL = FALSE, {
        ds()$set_schema(input$schema)
        tables <- ds()$get_tables()$table %||% list()
        shinyWidgets::updatePickerInput(session, "table", choices = tables)
      })

      observeEvent(input$table, ignoreNULL = FALSE, {
        shinycssloaders::showPageSpinner(
          ds()$set_table(input$table),
          caption = span("Reading table", tags$em(input$table))
        )
        nrows_large <- (!is.null(ds()$data) && nrow(ds()$data) >= DATASET_MAX_ROWS_TO_LOAD)
        shinyjs::toggle("too-many-rows-error", condition = nrows_large)
      })

      return(list(
        name = reactive( ds()$name ),
        data = reactive( ds()$data ),
        code = reactive( ds()$code ),
        error = reactive( ds()$error )
      ))
    }
  )
}

##########

SnowflakeDataSource <- R6::R6Class(
  "SnowflakeDataSource",
  inherit = DatabaseDataSource,

  private = list(
    get_full_info = function() {
      client <- ddr_function("datasource_client")()
      sql <- "SHOW TABLES"
      info <- as.data.frame(ddr_function("query")(client, self$dds$name, sql))
      info <- subset(info, select = c("database_name", "schema_name", "name", "rows"))
      names(info) <- c("database", "schema", "table", "rows")
      info$rows <- round(info$rows)
      if (self$fixed_database) {
        info$database <- self$database
      }
      if (self$fixed_schema) {
        info$schema <- self$schema
      }
      info
    }
  )
)
SnowflakeDataSource$shiny <- list(
  ui = function(id) {
    ns <- NS(id)

    tagList(
      shinyjs::hidden(
        shinyWidgets::alert(id = ns("no-warehouse-error"), "Snowflake data sources must have a warehouse set.", status = "warning")
      ),
      database_datasource_ui(ns("ds_snowflake"))
    )
  },

  server = function(id, ds) { moduleServer(id, function(input, output, session) {
    if (is.null(ds$dds$config$warehouse)) {
      shinyjs::show("no-warehouse-error")
      return()
    }

    database_datasource_server("ds_snowflake", ds = ds)
  })}
)

##########

PostgreSQLDataSource <- R6::R6Class(
  "PostgreSQLDataSource",
  inherit = DatabaseDataSource,

  private = list(
    get_full_info = function() {
      client <- ddr_function("datasource_client")()
      sql <- paste(
        "SELECT",
        "pgClass.relname::varchar AS name, pgNamespace.nspname::varchar AS schema_name, pgClass.reltuples AS rows",
        "FROM",
        "pg_class pgClass",
        "INNER JOIN",
        "pg_namespace pgNamespace ON (pgNamespace.oid = pgClass.relnamespace)",
        "WHERE",
        "pgNamespace.nspname NOT IN ('pg_catalog', 'pg_internal', 'information_schema') AND pgClass.relkind='r'"
      )
      info <- as.data.frame(ddr_function("query")(client, self$dds$name, sql))
      info <- subset(info, select = c("schema_name", "name", "rows"))
      names(info) <- c("schema", "table", "rows")
      info$rows <- round(info$rows)
      info$database <- self$database
      info
    }
  )
)
PostgreSQLDataSource$shiny <- list(
  ui = function(id) {
    ns <- NS(id)
    database_datasource_ui(ns("ds_postgresql"))
  },

  server = function(id, ds) { moduleServer(id, function(input, output, session) {
    database_datasource_server("ds_postgresql", ds = ds)
  })}
)

##########

RedshiftDataSource <- R6::R6Class(
  "RedshiftDataSource",
  inherit = PostgreSQLDataSource
)
RedshiftDataSource$shiny <- list(
  ui = function(id) {
    ns <- NS(id)
    database_datasource_ui(ns("ds_redshift"))
  },

  server = function(id, ds) { moduleServer(id, function(input, output, session) {
    database_datasource_server("ds_redshift", ds = ds)
  })}
)

##########

SQLServerDataSource <- R6::R6Class(
  "SQLServerDataSource",
  inherit = DatabaseDataSource,

  private = list(
    .limit_type = "top",

    get_full_info = function() {
      client <- ddr_function("datasource_client")()
      sql <- paste(
        "SELECT",
        "SCHEMA_NAME(t.uid) schema_name, t.name table_name, i.rows rows",
        "FROM",
        "sysobjects t, sysindexes i",
        "WHERE",
        "t.xtype = 'U' AND i.id = t.id AND i.indid IN (0,1)"
      )
      info <- as.data.frame(ddr_function("query")(client, self$dds$name, sql))
      names(info) <- c("schema", "table", "rows")
      info$rows <- round(info$rows)
      info$database <- self$database
      info
    }
  )
)
SQLServerDataSource$shiny <- list(
  ui = function(id) {
    ns <- NS(id)
    database_datasource_ui(ns("ds_sqlserver"))
  },

  server = function(id, ds) { moduleServer(id, function(input, output, session) {
    database_datasource_server("ds_sqlserver", ds = ds)
  })}
)

##########

MySQLDataSource <- R6::R6Class(
  "MySQLDataSource",
  inherit = DatabaseDataSource,

  private = list(
    .uses_schema = FALSE,

    get_full_info = function() {
      client <- ddr_function("datasource_client")()
      sql <- "SHOW TABLE STATUS"
      info <- ddr_function("query")(client, self$dds$name, sql)
      info$Rows <- info$Rows$cast(arrow::int64())
      info <- as.data.frame(info)
      info <- subset(info, select = c("Name", "Rows"))
      names(info) <- c("table", "rows")
      info$rows <- round(info$rows)
      info$database <- self$database
      info
    }
  )
)
MySQLDataSource$shiny <- list(
  ui = function(id) {
    ns <- NS(id)
    database_datasource_ui(ns("ds_mysql"))
  },

  server = function(id, ds) { moduleServer(id, function(input, output, session) {
    database_datasource_server("ds_mysql", ds = ds)
  })}
)

##########

OracleDataSource <- R6::R6Class(
  "OracleDataSource",
  inherit = DatabaseDataSource,

  private = list(
    .limit_type = "fetch",

    get_full_info = function() {
      client <- ddr_function("datasource_client")()
      sql <- paste(
        "SELECT",
        "OWNER AS \"schema\", TABLE_NAME AS \"table\", NUM_ROWS AS \"rows\"",
        "FROM",
        "ALL_TABLES",
        "WHERE",
        "OWNER NOT IN ('SYS', 'OUTLN', 'SYSTEM', 'DBSFWUSER', 'DBSNMP', 'APPQOSSYS', 'GSMADMIN_INTERNAL', 'XDB', 'CTXSYS', 'RDSADMIN', 'AUDSYS')"
      )
      info <- as.data.frame(ddr_function("query")(client, self$dds$name, sql))
      names(info) <- c("schema", "table", "rows")
      info$rows <- round(info$rows)
      info$database <- self$database
      info
    }
  )
)
OracleDataSource$shiny <- list(
  ui = function(id) {
    ns <- NS(id)
    database_datasource_ui(ns("ds_oracle"))
  },

  server = function(id, ds) { moduleServer(id, function(input, output, session) {
    database_datasource_server("ds_oracle", ds = ds)
  })}
)
