#' A subclass of LoadComponentDataSource that represents an object store data source
#'
#' @description
#' Object store data sources include services such as S3 and Google Cloud Storage
#'
#' @field file The path of the file (key from object store)
ObjectStoreDataSource <- R6::R6Class(
  "ObjectStoreDataSource",
  inherit = LoadComponentDataSource,

  private = list(
    .file = NULL,
    .params = NULL,

    get_name = function() {
      make.names(get_file_name_no_ext(self$file))
    },

    get_code = function() {
      filename <- basename(self$file)

      code <- glue::glue(
        'client <- { ddr_function("datasource_client", text = TRUE) }()',
        'temp <- { shQuote(filename, type = "cmd") }',
        '{ ddr_function("save_object", text = TRUE) }(client, { shQuote(self$dds$name, type = "cmd") }, { shQuote(self$file, type = "cmd") }, temp)',
        'rm(client)',
        code_load_file_by_file_type(filename, private$.params),
        .sep = '\n'
      )
      code
    },

    get_data = function() {
      suppressWarnings(eval(parse(text = private$.code)))
    }
  ),

  active = list(
    file = function() {
      private$.file
    }
  ),

  public = list(

    #' @description
    #' Get a list of files (keys) in the object store. Only txt and csv files are supported.
    get_files = function() {
      client <- ddr_function("datasource_client")()
      files <- ddr_function("list_keys")(client, self$dds$name, page_size = 5000)
      supported_files <- files[tools::file_ext(files) %in% LoadComponentFile$FILE_READ_EXTENSIONS]
      supported_files
    },

    #' @description
    #' Load a file from the object store
    #'
    #' @param file The path of the file (key from object store)
    #' @param params Named list of parameters to use for the function that loads the file
    set_file = function(file, params = list()) {
      if (!is.null(file) && file == "") {
        file <- NULL
      }
      private$.file <- file
      private$.params <- params

      if (is.null(file)) {
        private$.name <- private$.code <- private$.data <- private$.error <- NULL
      } else {
        private$run()
      }

      private$.invalidate()

      invisible(self)
    }
  )
)

#' Shiny module: Object store data source
#'
#' @param id Unique ID for the module
#' @param ds An ObjectStoreDataSource object
#' @return List of reactives:
#'   - name
#'   - data
#'   - code
#'   - error
#' @name object_store_datasource
NULL

#' @rdname object_store_datasource
object_store_datasource_ui <- function(id) {
  ns <- NS(id)

  tagList(
    uiOutput(ns("error")),
    shinyjs::hidden(
      div(
        id = ns("files-container"),
        shinyfilebrowser::list_selector_ui(ns("files"), bigger = TRUE),
        load_file_params_ui(ns("params"))
      )
    )
  )
}

#' @rdname object_store_datasource
object_store_datasource_server <- function(id, ds) {
  if (!inherits(ds, ObjectStoreDataSource$classname)) {
    stop("object_store_datasource_server: Only <ObjectStoreDataSource> objects are supported")
  }

  moduleServer(
    id,
    function(input, output, session) {

      files <- shinycssloaders::showPageSpinner(
        try(ds$get_files(), silent = TRUE),
        caption = span("Getting files from", tags$em(ds$dds$name))
      )

      if (inherits(files, "try-error")) {
        error <- attr(files, "condition")$message
        if (grepl("KeyError", error)) {
          error <- paste(error, "(This may be due to an incompatibility between your Domino version and the python dominodatalab-data module version)")
        }
        output$error <- renderUI({
          div(class = "alert alert-danger", icon("exclamation-sign", lib = "glyphicon"), error)
        })
        return()
      }

      selected_file <- shinyfilebrowser::list_selector_server(
        "files", choices = files, text_empty = paste("This", ds$dds$type, "bucket has no supported files")
      )
      params <- load_file_params_server("params", file_path = selected_file)
      shinyjs::show("files-container")

      ds <- ds$reactive()

      observeEvent(c(selected_file(), params()), ignoreNULL = FALSE, {
        shinycssloaders::showPageSpinner(
          ds()$set_file(selected_file(), params()),
          caption = span("Reading file", tags$em(selected_file()))
        )
      })

      return(list(
        name = reactive( ds()$name ),
        data = reactive( ds()$data ),
        code = reactive( ds()$code ),
        error = reactive( ds()$error )
      ))
    }
  )
}

S3DataSource <- R6::R6Class(
  "S3DataSource",
  inherit = ObjectStoreDataSource
)
S3DataSource$shiny <- list(
  ui = function(id) {
    ns <- NS(id)
    object_store_datasource_ui(ns("ds_s3"))
  },

  server = function(id, ds) { moduleServer(id, function(input, output, session) {
    object_store_datasource_server("ds_s3", ds = ds)
  })}
)

GCSDataSource <- R6::R6Class(
  "GCSDataSource",
  inherit = ObjectStoreDataSource
)
GCSDataSource$shiny <- list(
  ui = function(id) {
    ns <- NS(id)
    object_store_datasource_ui(ns("ds_gcs"))
  },

  server = function(id, ds) { moduleServer(id, function(input, output, session) {
    object_store_datasource_server("ds_gcs", ds = ds)
  })}
)

ADLSDataSource <- R6::R6Class(
  "ADLSDataSource",
  inherit = ObjectStoreDataSource
)
ADLSDataSource$shiny <- list(
  ui = function(id) {
    ns <- NS(id)
    object_store_datasource_ui(ns("ds_adls"))
  },

  server = function(id, ds) { moduleServer(id, function(input, output, session) {
    object_store_datasource_server("ds_adls", ds = ds)
  })}
)

GenericS3DataSource <- R6::R6Class(
  "GenericS3DataSource",
  inherit = ObjectStoreDataSource
)
GenericS3DataSource$shiny <- list(
  ui = function(id) {
    ns <- NS(id)
    object_store_datasource_ui(ns("ds_genericS3"))
  },

  server = function(id, ds) { moduleServer(id, function(input, output, session) {
    object_store_datasource_server("ds_genericS3", ds = ds)
  })}
)
