#' R6 superclass for loading datasets
#'
#' @description
#' A subclass must implement:
#'   - `private$get_name()`
#'   - `private$get_code()`
#'   - `private$get_data()`
#'
#' @field name The name of the dataset (or `NULL` if there's an error)
#' @field code The code to load the dataset (or `NULL` if there's an error)
#' @field data The dataframe (or `NULL` if there's an error)
#' @field error An error message generated while loading the dataset (or `NULL`)
LoadComponent <- R6::R6Class(
  "LoadComponent",
  cloneable = FALSE,

  private = list(
    .name = NULL,
    .code = NULL,
    .data = NULL,
    .error = NULL,

    run = function() {
      tryCatch({
        private$.name <- private$get_name()
        private$.code <- private$get_code()
        private$.data <- private$get_data()
        private$.error <- NULL

        if (!is.data.frame(private$.data)) {
          stop("The resulting object must be a data.frame, not <", class(private$.data)[1], ">")
        }
        if (nrow(private$.data) == 0) {
          stop("The resulting data.frame has no rows")
        }
      }, error = function(err) {
        private$.name <- private$.code <- private$.data <- NULL
        private$.error <- glue::glue("Could not load data ({ err$message })")
      })
      invisible(self)
    }
  ),

  active = list(
    name = function() private$.name,
    code = function() private$.code,
    data = function() private$.data,
    error = function() private$.error
  ),

  public = list(
    #' @description
    #' Create a new instance
    initialize = function() {
      if (is.null(private$get_name) || is.null(private$get_code) || is.null(private$get_data)) {
        stop(class(self)[1], ": The following private functions must be implemented: `get_name()`, `get_code()`, `get_data()`.")
      }
      invisible(self)
    },

    #' @description
    #' Print a string representation
    print = function() {
      if (!is.null(private$.error)) {
        cat(glue::glue("<LoadComponent> Error: { private$.error }\n"))
      } else {
        cat("<LoadComponent>\n")
        cat0("Name: ", private$.name, "\n")
        cat0("Code: ", private$.code, "\n")
        str(private$.data)
      }
    }
  )
)
