#' A subclass of LoadComponent that loads data from a DominoDataSource
#'
#' @description
#' A subclass must call `private$run()` and implement:
#'   - `private$get_name()`
#'   - `private$get_code()`
#'   - `private$get_data()`
#'
#' @field dds The DominoDataSource associated with this DataSource
LoadComponentDataSource <- R6::R6Class(
  "LoadComponentDataSource",
  inherit = LoadComponent,

  private = list(
    .dds = NULL,

    # variables to allow this class to be reactive
    .rx_dep = NULL,
    .rx_expr = NULL,
    .rx_count = 0,
    .invalidate = function() {
      private$.rx_count <- private$.rx_count + 1
      private$.rx_dep(private$.rx_count)
      invisible()
    }
  ),

  active = list(
    dds = function() {
      private$.dds
    }
  ),

  public = list(

    #' @description
    #' Create a new instance
    #'
    #' @param dds The DominoDataSource associated with this DataSource
    initialize = function(dds) {
      super$initialize()
      private$.dds <- dds
      private$.rx_dep <- function(x) NULL
      invisible(self)
    },

    #' @description
    #' Print a string representation
    print = function() {
      cat(glue::glue("<{ class(self)[1]}> "))
      self$dds$print()
    },

    #' @description
    #' Get a reactive version of this object that invalidates when it changes
    reactive = function() {
      # Idea borrowed from Winston Chang
      # https://community.rstudio.com/t/good-way-to-create-a-reactive-aware-r6-class
      if (is.null(private$.rx_expr)) {
        private$.rx_dep <- shiny::reactiveVal(0)
        private$.rx_expr <- shiny::reactive({
          private$.rx_dep()
          self
        })
      }
      private$.rx_expr
    }
  )
)

#' Shiny module: Load data from a Domino DataSource
#'
#' @param id Unique ID for the module
#' @return List of reactives:
#'   - name
#'   - data
#'   - code
#'   - error
#' @name LoadComponentDataSourceShiny
LoadComponentDataSource$shiny <- list(

  ui = function(id) {
    ns <- NS(id)

    if (!is_data_sources_supported()) {
      return(
        tagList(
          br(),
          shinyWidgets::alert(id = ns("not-supported"), glue::glue("Data Sources are supported on Domino version { DATASOURCES_MIN_VERSION } and above."), status = "warning")
        )
      )
    }

    tagList(
      html_dependency_dca(),
      shinyjs::useShinyjs(),
      br(),
      shinyjs::hidden(
        shinyWidgets::pickerInput(ns("data_source"), "Data Source", choices = NULL, multiple = TRUE, width = "100%",
                                  options = list(`max-options` = 1, `none-selected-text` = "Select a Data Source"))
      ),
      div(id = ns("datasource_ui"))
    )
  },

  server = function(id) {
    moduleServer(
      id,
      function(input, output, session) {
        if (!is_data_sources_supported()) {
          return(list(
            name = reactive(NULL),
            data = reactive(NULL),
            code = reactive(NULL),
            error = reactive(NULL)
          ))
        }

        all_data_sources_unsorted <- shinycssloaders::showPageSpinner(
          get_data_sources(),
          caption = "Loading data sources"
        )

        all_data_sources <- c(
          Filter(function(dds) is_data_source_supported(dds), all_data_sources_unsorted),
          Filter(function(dds) !is_data_source_supported(dds), all_data_sources_unsorted)
        )

        if (length(all_data_sources) == 0) {
          ui <- shinyWidgets::alert("This project does not have any Data Sources available.", status = "warning")
          insertUI(paste0("#", session$ns("datasource_ui")), "afterBegin", ui)
          return(list(
            name = reactive(NULL),
            data = reactive(NULL),
            code = reactive(NULL),
            error = reactive(NULL)
          ))
        } else {
          shinyjs::show("data_source")
          choice_values <- sapply(all_data_sources, function(dds) dds$name)
          choice_html <- sapply(all_data_sources, get_data_source_dropdown_html)
          choice_disabled <- sapply(all_data_sources, function(dds) !is_data_source_supported(dds))
          shinyWidgets::updatePickerInput(session, "data_source", choices = choice_values,
                                          choicesOpt = list(content = choice_html, disabled = choice_disabled))
        }

        selected_source <- reactive({
          req(input$data_source)
          Filter(function(dds) dds$name == input$data_source, all_data_sources)[[1]]
        })
        selected_source_class <- reactive({
          req(selected_source())
          ds_class <- paste0(selected_source()$type, "DataSource")
          get(ds_class)
        })

        new_ds_added <- reactiveVal(0)
        module_id <- reactive({
          req(selected_source_class())
          paste0("ds_", new_ds_added())
        })
        observe({
          removeUI(paste0("#", session$ns("datasource_ui"), " > .data-source-container"), immediate = TRUE)
          req(selected_source_class())
          isolate(new_ds_added(new_ds_added() + 1))
          id <- session$ns(module_id())
          ui <- div(
            id = id,
            class = "data-source-container",
            selected_source_class()$shiny$ui(id)
          )
          insertUI(paste0("#", session$ns("datasource_ui")), "afterBegin", immediate = TRUE, ui)
        })
        ds_module <- reactive({
          new_ds_added()
          isolate({
            dds <- selected_source()
            ds <- selected_source_class()$new(dds = dds)
            selected_source_class()$shiny$server(module_id(), ds)
          })
        })

        return(list(
          name = reactive(tryCatch(ds_module()$name(), error = function(err) NULL)),
          data = reactive(tryCatch(ds_module()$data(), error = function(err) NULL)),
          code = reactive(tryCatch(ds_module()$code(), error = function(err) NULL)),
          error = reactive(tryCatch(ds_module()$error(), error = function(err) NULL))
        ))

      }
    )
  }
)
