#' Call a function from {DominoDataR} or from its mock
#'
#' If an environment variable `MOCK_DOMINODATAR` exists, then mock functions will be used.
#' Otherwise, the functions from {DominoDataR} will be used.
#' @param fxn The function name
#' @param text If `FALSE`, get the function. If `TRUE`, get a string that calls it.
ddr_function <- function(fxn, text = FALSE) {
  if (Sys.getenv("MOCK_DOMINODATAR") == "") {
    if (text) {
      paste0("DominoDataR::", fxn)
    } else {
      utils::getFromNamespace(fxn, "DominoDataR")
    }
  } else {
    fxn <- paste0("mock_ddr_", fxn)
    if (text) {
      fxn
    } else {
      utils::getFromNamespace(fxn, PACKAGE_NAME)
    }
  }
}

#' Mock function for DominoDataR::datasource_client
#'
#' The real function returns a DataSourceClient python object.\cr\cr
#' The mock function returns a directory path to where mock data sources are located.\cr\cr
#' Environment variable `MOCK_DOMINODATAR` must be set to a directory containing mock data sources.
#'
#' @param ... Parameters used for the real function but not in the mock function
mock_ddr_datasource_client <- function(...) {
  folder <- Sys.getenv("MOCK_DOMINODATAR")
  if (folder == "" || !dir.exists(folder)) {
    stop("Mock datasource_client: To use mock data sources, you must set `MOCK_DOMINODATAR` envvar to a path containing mock data.")
  }
  structure(folder, class = "MockDataSourceClient")
}

#' Mock function for DominoDataR::list_keys
#'
#' The real function returns the keys of an object store.\cr\cr
#' The mock function returns the file names inside a mock object store directory.
#'
#' @param client A MockDataSourceClient object returned from `mock_ddr_datasource_client()`
#' @param datasource The name of the data source (folder name)
#' @param ... Parameters used for the real function but not in the mock function
mock_ddr_list_keys <- function(client, datasource, ...) {
  folder <- get_mock_ddr_folder(client, datasource)
  list.files(folder, recursive = TRUE)
}

#' Mock function for DominoDataR::save_object
#'
#' The real function saves a file from an object store to local disk.\cr\cr
#' The mock function saves a file from a mock object store to local disk.
#'
#' @param client A MockDataSourceClient object returned from `mock_ddr_datasource_client()`
#' @param datasource The name of the data source (folder name)
#' @param object File from the mock data source
#' @param file Path to save the file
#' @param ... Parameters used for the real function but not in the mock function
mock_ddr_save_object <- function(client, datasource, object, file, ...) {
  folder <- get_mock_ddr_folder(client, datasource)
  from <- file.path(folder, object)
  if (!file.exists(from)) {
    stop("Mock save_object: object doesn't exist: ", from)
  }
  file.copy(from = from, to = file, overwrite = TRUE)
}

#' Mock function for DominoDataR::query
#'
#' The real function runs a SQL query against the given database.\cr\cr
#' The mock function can do two limited actions: if the query starts with `SELECT * FROM`
#' then the table name from the query is parsed out and assumed to be a CSV file in the mock
#' data source, and that CSV file is returned. Otherwise, the query is assumed to be asking
#' for a list of tables in the database (CSV files in the data source directory), so a
#' dataframe is returned with: database, schema, table, rows.
#'
#' @param client A MockDataSourceClient object returned from `mock_ddr_datasource_client()`
#' @param datasource The name of the data source (folder name)
#' @param query A SQL query
#' @param ... Parameters used for the real function but not in the mock function
mock_ddr_query <- function(client, datasource, query, ...) {
  folder <- get_mock_ddr_folder(client, datasource)

  if (startsWith(tolower(query), "select * from")) {
    table <- get_mock_ddr_table_name(query)
    file <- file.path(folder, paste0(table, ".csv"))
    if (!file.exists(file)) {
      stop("Mock query: `", table, "` table doesn't exist (", file, ")")
    }
    utils::read.csv(file)
  } else {
    tables <- list.files(folder, full.names = TRUE)
    rows <- unname(sapply(tables, function(x) nrow(utils::read.csv(x))))
    table_names <- tools::file_path_sans_ext(basename(tables))
    data.frame(database_name = "mockdb", schema_name = "mockschema", name = table_names, rows = rows)
  }
}

#' Get the folder holding a mock data source
#'
#' @param client A MockDataSourceClient object returned from `mock_ddr_datasource_client()`
#' @param datasource The name of the data source (folder name)
get_mock_ddr_folder <- function(client, datasource) {
  if (!inherits(client, "MockDataSourceClient")) {
    stop("Mock data sources: client must be the result of mock_ddr_datasource_client")
  }
  folder <- file.path(client, datasource)
  if (!dir.exists(folder)) {
    stop("Mock data sources: datasource folder doesn't exist: ", folder)
  }
  folder
}

#' Get the name of a table from a select query
#'
#' @param query A `SELECT * FROM` SQL query
get_mock_ddr_table_name <- function(query) {
  table_name <- sub("SELECT \\* FROM (\\S+).*", "\\1", query, ignore.case = TRUE)
  if (table_name == query) {
    stop("Could not find mock table name in query: ", query)
  }
  table_name <- utils::tail(strsplit(table_name, "\\.")[[1]], 1)
  table_name
}
