#' Shiny module: Select a dataframe from the global environment
#'
#' @param id Unique ID for the module
#' @param wizard The name of the wizard that is using this module
#' @param demo If `TRUE`, allow the user to choose a demo dataset if the user has
#' no available dataframes.
#' @return Clicking "Load more data" stops the app, runs the data loading wizard,
#' and re-launches the current wizard after a dataframe is created.
#' This module itself returns a list of reactives:
#'   - name: The name of the dataset (or `NULL`)
#'   - data: The dataframe (or `NULL`)
#'   - code: The code to load the dataset (or `NULL`)
#' @name data_environment
NULL

#' @rdname data_environment
data_environment_ui <- function(id) {
  ns <- NS(id)

  tagList(
    shinyjs::useShinyjs(),
    shinyWidgets::alert("Select a data.frame from your R workspace", status = "info"),
    shinyfilebrowser::list_selector_ui(ns("dataframes")),
    shinyjs::hidden(
      actionButton(ns("load_new"), "Load more data", icon("database"))
    )
  )
}

#' @rdname data_environment
data_environment_server <- function(id, wizard = NULL, demo = FALSE) {
  moduleServer(
    id,
    function(input, output, session) {

      DEMO_DF_NAME <- "df_quick_start"
      result <- reactiveValues(name = NULL, code = NULL, data = NULL)

      dataframes <- get_dataframes()
      df_dims <- get_df_dimensions(dataframes)
      if (length(dataframes) == 0) {
        if (demo) {
          df_choices <- DEMO_DF_NAME
        } else {
          df_choices <- c()
        }
      } else {
        df_choices <- stats::setNames(dataframes, paste(dataframes, df_dims))
      }
      datasets <- shinyfilebrowser::list_selector_server(
        "dataframes",
        df_choices,
        text_empty = "There are no data.frames"
      )

      observeEvent(datasets(), {
        if (length(dataframes) == 0 && datasets() == DEMO_DF_NAME) {
          assign_to_global(DEMO_DF_NAME, assistdomino::palmerpenguins)
          insert_text(glue::glue("{ DEMO_DF_NAME } <- { PACKAGE_NAME }::palmerpenguins"))
        }
      })

      name <- reactive({
        if (is.null(datasets())) {
          return()
        }
        datasets()
      })

      code <- reactive({
        if (is.null(datasets())) {
          return()
        }
        datasets()
      })

      observeEvent(code(), ignoreNULL = FALSE, {
        if (is.null(code())) {
          result$name <- NULL
          result$code <- NULL
          result$data <- NULL
        } else {
          data <- eval(parse(text = code()))
          result$name <- name()
          result$code <- code()
          result$data <- data
        }
      })

      if (!is.null(wizard)) { # nocov start
        shinyjs::show("load_new")

        observeEvent(input$load_new, {
          session$onSessionEnded(function() {
            code <- glue::glue("{ PACKAGE_NAME }:::dca_data({ shQuote(wizard) })")
            rstudioapi::sendToConsole(code)
          })
          kill_app()
        })
      } # nocov end

      return(list(
        name = reactive(result$name),
        data = reactive(result$data),
        code = reactive(result$code)
      ))
    }
  )
}

#' Get a list of all available dataframes in a user's workspace
#'
#' @param env The environment in which to look for dataframes (global env by default)
#' @return List of names of dataframes that exist
get_dataframes <- function(env = .GlobalEnv) {
  all_obj_names <- ls(envir = env)
  dfs <- lapply(all_obj_names, function(x) {
    obj <- get(x, envir = env)
    if (methods::is(obj, "data.frame")) {
      x
    } else {
      NULL
    }
  })
  dfs <- unlist(dfs)
  dfs
}

#' Get a human-friendly representation of the dimensions of dataframes
#'
#' @param dfs List of dataframe names
#' @param env The environment in which to look for dataframes (global env by default)
#' @return List with the same length as the input, showing the dimensions of each dataframe
get_df_dimensions <- function(dfs, env = .GlobalEnv) {
  lapply(dfs, function(df) {
    obj <- get(df, envir = env)
    glue::glue("({ natural_num(nrow(obj)) } rows x { natural_num(ncol(obj)) } cols)")
  })
}
