#' Shiny module: Snippets page
#'
#' @param id Unique ID for the module
#' @return List of reactives:
#'   - code: The lines of code from the selected snippet as a single string (or `""`)
#' @name page_snippets
NULL

#' @rdname page_snippets
page_snippets_ui <- function(id) {
  ns <- NS(id)

  tagList(
    use_mixpanel(),
    shinyjs::useShinyjs(),
    html_dependency_dca(),

    title_bar("Snippets"),

    div(
      class = "page-main-content",
      fluidRow(
        column(
          6,
          br(),
          shinyfilebrowser::path_browser_ui(ns("snippets"), bigger = TRUE)
        ),
        column(
          6,
          div(class = "snippets-file-name", textOutput(ns("file_name"))),
          shinycodeviewer::code_viewer_ui(ns("file_contents"))
        )
      )
    ),

    div(
      class = "page-actions flex flex-gap2",

      snippet_edit_mode_ui(ns("snippet-edit-buttons")),

      div(class = "flex-push"),
      actionButton(
        ns("close"),
        "Close",
        icon = icon("close"),
        class = "btn-lg"
      ),
      actionButton(
        ns("continue"),
        "Insert Code",
        icon = icon("check"),
        class = "btn-primary btn-lg"
      )
    )
  )
}

#' @rdname page_snippets
page_snippets_server <- function(id) {
  moduleServer(
    id,
    function(input, output, session) {
      shinymixpanel::mp_track(
        MIXPANEL_EVENT_INIT,
        list(
          section = MIXPANEL_SECTION_SNIPPETS
        )
      )

      observeEvent(input$close, {
        kill_app()
      })

      snippets_refresh <- reactive_trigger()

      all_snippets <- reactive({
        snippets_refresh$depend()
        merge_all_snippets()
      })

      snippet_selector <- shinyfilebrowser::path_browser_server(
        "snippets",
        reactive(sort(all_snippets()$short_path)),
        clear_selection_on_navigate = TRUE,
        show_extension = FALSE,
        text_empty = "No snippets here"
      )

      selected_snippet_short <- reactive({
        snippet_selector$selected()
      })

      selected_snippet <- reactive({
        if (is.null(selected_snippet_short())) {
          return()
        }
        all_snippets()[all_snippets()$short_path == selected_snippet_short(), ]$full_path
      })

      output$file_name <- renderText({
        if (is.null(selected_snippet())) {
          "Please select a snippet to preview"
        } else {
          get_file_name_no_ext(selected_snippet())
        }
      })

      file_contents <- reactive({
        if (is.null(selected_snippet())) {
          NULL
        } else {
          lines <- suppressWarnings(readLines(selected_snippet()))
          lines
        }
      })
      file_contents_string <- reactive({
        paste(file_contents(), collapse = "\n")
      })

      shinycodeviewer::code_viewer_server("file_contents", chunks = file_contents, editable = FALSE, show_chunk_numbers = TRUE)

      snippet_edit_mode_server(
        "snippet-edit-buttons",
        selected_snippet = selected_snippet,
        curwd = reactive(snippet_selector$path()),
        snippets_refresh = snippets_refresh
      )

      observe({
        shinyjs::toggleState("continue", condition = !is.null(selected_snippet()))
      })

      observeEvent(input$continue, {
        insert_text(file_contents_string(), remove_duplicate_libs = FALSE)

        shinymixpanel::mp_track(
          MIXPANEL_EVENT_CODE,
          list(
            section = MIXPANEL_SECTION_SNIPPETS
          )
        )

        kill_app()
      })

      return(
        list(
          code = file_contents_string
        )
      )
    }
  )
}
