#' Shiny module: Select arbitrary number of pairs from two dropdown menus
#'
#' Create two dropdown menus and allow the user to choose any number of pairs from these
#' dropdowns. Selecting duplicate pairs is not allowed.
#'
#' Technical implementation detail: a hidden text field saves the existing pairs
#' as a JSON list-of-lists (a single list can't be used because in JavaScript a dictionary
#' cannot have duplicate keys).
#'
#' @param id Unique ID for the module
#' @param label_1 Label for first dropdown
#' @param label_2 Label for second dropdown
#' @param choices_1 List of choices for the first dropdown. Leave as `NULL` and use `choices_df_1`
#' instead if the choices should be the column names of some dataframe.
#' @param choices_2 List of choices for the second dropdown. Leave as `NULL` and use `choices_df_2`
#' instead if the choices should be the column names of some dataframe.
#' @param choices_df_1 (reactive) A dataframe whose column names become the list of choices.
#' @param choices_df_2 (reactive) A dataframe whose column names become the list of choices.
#' @return List:
#'   - get: (reactive) Named list containing all the selected pairs.
#'   - set: A function accepting a named list that sets the UI to contain the given list of pairs.
#' @name pairs_selector
NULL

#' @rdname pairs_selector
pairs_selector_ui <- function(id, label_1 = NULL, label_2 = NULL, choices_1 = NULL, choices_2 = NULL) {
  ns <- NS(id)

  tagList(
    html_dependency_dca(),
    shinyjs::useShinyjs(),
    div(
      class = "pair_selector",
      div(
        class = "pair_selector_header",
        shinyWidgets::pickerInput(ns("choice_1"), label_1, choices_1, multiple = TRUE, options = list(`max-options` = 1)),
        shinyWidgets::pickerInput(ns("choice_2"), label_2, choices_2, multiple = TRUE, options = list(`max-options` = 1)),
        tags$button(icon("times"), class = "btn space-maker")
      ),
      uiOutput(ns("existing_pairs_ui"), class = "existing_pairs"),
      shinyjs::hidden(textInput(ns("existing_pairs_raw"), "", "[]"))
    )
  )
}

#' @rdname pairs_selector
pairs_selector_server <- function(id, choices_df_1 = reactive(NULL), choices_df_2 = reactive(NULL)) {
  moduleServer(
    id,
    function(input, output, session) {
      observeEvent(choices_df_1(), {
        req(choices_df_1())

        new_choices <- names(choices_df_1())
        shinyWidgets::updatePickerInput(session, "choice_1", choices = new_choices)

        if (length(existing_pairs()) > 0) {
          existing_pairs_names <- names(unlist(existing_pairs()))
          if (!all(existing_pairs_names %in% new_choices)) {
            updateTextInput(session, "existing_pairs_raw", value = "[]")
          }
        }
      })
      observeEvent(choices_df_2(), {
        req(choices_df_2())

        new_choices <- names(choices_df_2())
        shinyWidgets::updatePickerInput(session, "choice_2", choices = new_choices)

        if (length(existing_pairs()) > 0) {
          existing_pairs_names <- unname(unlist(existing_pairs()))
          if (!all(existing_pairs_names %in% new_choices)) {
            updateTextInput(session, "existing_pairs_raw", value = "[]")
          }
        }
      })

      observeEvent(c(input$choice_1, input$choice_2), {
        req(input$choice_1, input$choice_2)
        shinyWidgets::updatePickerInput(session, "choice_1", selected = "")
        shinyWidgets::updatePickerInput(session, "choice_2", selected = "")

        new_pair <- list(as.list(stats::setNames(input$choice_2, input$choice_1)))
        if (new_pair %in% existing_pairs()) {
          return()
        }
        new_pairs <- append(existing_pairs(), new_pair)
        updateTextInput(session, "existing_pairs_raw", value = as.character(jsonlite::toJSON(new_pairs)))
      })

      existing_pairs <- reactive({
        if (is.null(input$existing_pairs_raw)) {
          list()
        } else {
          jsonlite::fromJSON(input$existing_pairs_raw, simplifyDataFrame = FALSE)
        }
      })

      output$existing_pairs_ui <- renderUI({
        lapply(seq_along(existing_pairs()), function(idx) {
          div(
            class = "pair_row",
            tags$input(
              value = names(existing_pairs()[[idx]][1]),
              type="text", disabled=NA, class="form-control"
            ),
            tags$input(
              value = existing_pairs()[[idx]][[1]],
              type="text", disabled=NA, class="form-control"
            ),
            tags$button(
              icon("times"), class = "remove-btn btn", title = "Remove",
              onclick = glue::glue("Shiny.setInputValue('{{session$ns('remove_pair')}}', {{idx}}, {priority: 'event'})", .open = "{{", .close = "}}")
            )
          )
        })
      })

      observeEvent(input$remove_pair, {
        pairs <- existing_pairs()
        pairs[input$remove_pair] <- NULL
        updateTextInput(session, "existing_pairs_raw", value = as.character(jsonlite::toJSON(pairs)))
      })

      set_pairs <- function(pairs = list()) {
        pairs_list_of_lists <- lapply(seq_along(pairs), function(idx) {
          as.list(pairs[idx])
        })
        updateTextInput(session, "existing_pairs_raw", value = as.character(jsonlite::toJSON(pairs_list_of_lists)))
      }

      return(list(
        get = reactive(unlist(existing_pairs())),
        set = set_pairs
      ))
    }
  )
}
