# nocov start

#' Domino Code Assist - Data selection
#'
#' Run the data selection DCA wizard.
#' @param next_wizard The next wizard to call after the data selection wizard completes.
#' @export
dca_data <- function(next_wizard = NULL) {
  app <- shiny::shinyApp(
    ui = shiny::fluidPage(page_data_select_ui("app")),
    server = function(input, output, session) page_data_select_server("app")
  )
  ret <- shiny::runGadget(app, viewer = domino_viewer())

  if (!is.null(next_wizard) && !is.null(ret)) {
    code <- glue::glue("{ PACKAGE_NAME }:::dca_{ next_wizard }({ shQuote(ret) })")
    # using sendToConsole is a workaround for https://github.com/rstudio/rstudio/issues/12760
    rstudioapi::sendToConsole(code)
  }
}

#' Domino Code Assist - Data selection
#'
#' Run the data selection DCA wizard.
#' @export
dca_data_addin <- function() {
  dca_data()
}

#' Domino Code Assist - Transformations
#'
#' Run the data transformation DCA wizard.
#' @param data_name The name of a data frame in the global environment. Can be reactive or static.
#' @export
dca_transform <- function(data_name = NULL) {
  app <- shiny::shinyApp(
    ui = shiny::fluidPage(page_xforms_ui("app")),
    server = function(input, output, session) page_xforms_server("app", data_name_in = data_name)
  )
  shiny::runGadget(app, viewer = domino_viewer())
}

#' Domino Code Assist - Transformations
#'
#' Run the data transformation DCA wizard. If a variable holding a data frame is selected
#' when invoking the addin in RStudio, that data frame will be used.
#' @export
dca_transform_addin <- function() {
  data_name <- tryCatch({
    context <- rstudioapi::getActiveDocumentContext()
    name <- context$selection[[1]]$text

    if (nzchar(name)) {
      name
    } else {
      NULL
    }
  }, error = function(err) {
    NULL
  })

  dca_transform(data_name = data_name)
}

#' Domino Code Assist - Visualizations
#'
#' Run the data visualizations DCA wizard.
#' @param data_name The name of a data frame in the global environment. Can be reactive or static.
#' @export
dca_viz <- function(data_name = NULL) {
  app <- shiny::shinyApp(
    ui = shiny::fluidPage(page_viz_ui("app")),
    server = function(input, output, session) page_viz_server("app", data_name_in = data_name)
  )
  shiny::runGadget(app, viewer = domino_viewer())
}

#' Domino Code Assist - Visualizations
#'
#' Run the data visualizations DCA wizard. If a variable holding a data frame is selected
#' when invoking the addin in RStudio, that data frame will be used.
#' @export
dca_viz_addin <- function() {
  data_name <- tryCatch({
    context <- rstudioapi::getActiveDocumentContext()
    name <- context$selection[[1]]$text

    if (nzchar(name)) {
      name
    } else {
      NULL
    }
  }, error = function(err) {
    NULL
  })

  dca_viz(data_name = data_name)
}

#' Domino Code Assist - Snippets
#'
#' Run the snippets DCA wizard.
#' @export
dca_snippets <- function() {
  app <- shiny::shinyApp(
    ui = shiny::fluidPage(page_snippets_ui("app")),
    server = function(input, output, session) page_snippets_server("app")
  )
  shiny::runGadget(app, viewer = domino_viewer())
}

#' Domino Code Assist - Snippets
#'
#' Run the snippets DCA wizard.
#' @export
dca_snippets_addin <- function() {
  dca_snippets()
}

domino_viewer <- function() {
  title <- paste0("Domino Code Assist", " v", as.character(utils::packageVersion(PACKAGE_NAME)))
  shiny::dialogViewer(title, 1100, 800)
}

# nocov end
