#' Create the UI for the title of a page
#'
#' @param title The title of the page
title_bar <- function(title) {
  shiny::tagList(
    shiny::tags$html(class = "dca-page"),
    shiny::div(
      html_dependency_dca(),
      class = "header-bar",
      class = "flex flex-gap1",
      title,
      shiny::div("Domino Code Assist", class = "flex-push"),
      shiny::img(height = "100%", src = "dca-assets/dca/img/domino-logo.svg")
    )
  )
}

#' Create an htmldependency for DCA
html_dependency_dca <- function() {
  htmltools::htmlDependency(
    name = "dca",
    version = utils::packageVersion(PACKAGE_NAME),
    src = c(href = "dca-assets/dca"),
    package = PACKAGE_NAME,
    script = c("js/shiny-utils.js"),
    stylesheet = c("css/dca.css")
  )
}

#' Reactive triggger to explicitly trigger a reactive expression
#'
#' @return A list of two functions:
#'   - depend: Call this to create a reactive dependency
#'   - trigger: Call this to trigger an invalidation, thus all dependencies will run
reactive_trigger <- function() {
  rv <- shiny::reactiveVal(0)
  list(
    depend = function() {
      invisible(rv())
    },
    trigger = function() {
      rv(isolate(rv() + 1))
    }
  )
}

#' A reactiveVal object that invalidates every time its value is set
#'
#' Regular reactiveVal only invalidate when their values change. This construct
#' will invalidate every time its value is set, even if the new value is the
#' same as the old value
#'
#' @param data The initial value
#' @return Similarly to reactiveVal: a function that reads the value when called with
#' no arguments, or sets the value if called with a single argument.
reactiveValEvent <- function(data = NULL) {
  rv <- shiny::reactiveVal(data)
  rv_trigger <- shiny::reactiveVal(FALSE)
  ret <- function(data) {
    if (missing(data)) {
      rv_trigger() # Take a reactive dependency on rv_trigger
      rv()
    } else {
      rv_trigger(shiny::isolate(!rv_trigger())) # This should always result in rv_trigger invalidation
      rv(data)
    }
  }
  class(ret) <- class(rv) # Make sure this object will have "reactive" class
  ret
}

#' Stop a shiny app and close its browser tab
#'
#' @param value The value that should be returned from the Shiny app
#' @param session The shiny session to close
kill_app <- function(value = invisible(), session = shiny::getDefaultReactiveDomain()) {
  session$sendCustomMessage("dca-close-window", TRUE)
  shiny::stopApp(value)
}

#' Return a reactive version of a variable
#'
#' @param x Any object, either reactive or static
#' @return A reactive version of the input
make_reactive <- function(x) {
  if (shiny::is.reactive(x)) {
    x
  } else {
    shiny::reactive(x)
  }
}

#' Add a checkbox next to an existing input
#'
#' @param input A shiny input UI
#' @param checkboxId The ID of the checkbox to create
#' @param checkboxLabel The label of the checkbox to create
#' @param checkboxValue The value of the checkbox to create (TRUE or FALSE)
#' @param side Where to place the checkbox in relation to the input ("left" or "right")
#' @return UI for the input and checkbox together
input_with_checkbox <- function(input, checkboxId, checkboxLabel, checkboxValue = FALSE, side = c("left", "right")) {
  stopifnot(inherits(input, "shiny.tag"))
  side <- match.arg(side)

  checkbox <- tags$label(
    class = "input-group-addon",
    tags$input(
      id = checkboxId,
      checked = if (isTRUE(checkboxValue)) "checked" else NULL,
      type = "checkbox", style = "vertical-align: middle"
    ),
    tags$span(checkboxLabel, style = "vertical-align: middle")
  )

  input_idx <- try(which(sapply(input$children, function(x) x$name == "input")), silent = TRUE)
  if (inherits(input_idx, "try-error") || length(input_idx) != 1) {
    stop("Can't find the <input> tag")
  }

  input_tag <- input$children[[input_idx]]

  if (side == "left") {
    group <- tags$div(class = "input-group", checkbox, input_tag)
  } else if (side == "right") {
    group <- tags$div(class = "input-group", input_tag, checkbox)
  } else {
    stop("Invalid 'side'")  # nocov
  }

  input$children[[input_idx]] <- group
  input
}
