#' Get directories of builtin snippets
get_snippets_paths_builtin <- function() {
  path <- system.file("sample_snippets", "snippets", package = PACKAGE_NAME)
  if (dir.exists(path)) {
    path
  } else {
    character(0)
  }
}

#' Get directories of snippets in git repositories
get_snippets_paths_git <- function() {
  Sys.glob(file.path(get_user_git_dir(),"*", "snippets"))
}

#' Get directories of snippets in imported Domino projects
get_snippets_paths_imported_projects <- function() {
  file.path(get_imported_projects_dirs(), "snippets")
}

#' Get directories of snippets in the current project
get_snippets_paths_current_project <- function() {
  file.path(get_user_project_dir(), "snippets")
}

#' Find all snippets in a given directory (recursively)
#'
#' Snippets are any file ending in .R extension under the given path.
#'
#' @param path Directory where to search for snippets. Must end in a directory
#' named "snippets".
#' @return A dataframe where each row represents a snippet two columns:
#'   - full_path: The full normalized path to the snippet
#'   - short_path: A path to the snippet starting at the input path
find_snippets_path <- function(path) {
  empty_df <- data.frame(full_path = character(0), short_path = character(0))
  if (!dir.exists(path)) {
    return(empty_df)
  }
  if (basename(path) != "snippets") {
    return(empty_df)
  }
  data.frame(
    full_path = normalizePath(winslash = "/", list.files(path, recursive = TRUE, pattern = "\\.R$", full.names = TRUE)),
    short_path = list.files(path, recursive = TRUE, pattern = "\\.R$", full.names = FALSE),
    stringsAsFactors = FALSE
  )
}

#' Find all snippets in a list of paths
#'
#' If multiple snippets have the same `short_path` (for example, there can be
#' two input paths that both have a file with the same name), only the first one
#' is kept. This means that the `paths[n]` (from the input) takes priority over
#' `paths[n+1]`.
#'
#' @param paths List of paths to search for snippets
#' @return A dataframe where each row represents a snippet two columns:
#'   - full_path: The full normalized path to the snippet
#'   - short_path: A path to the snippet starting at the input path
find_snippets <- function(paths) {
  dfs <- lapply(paths, find_snippets_path)
  df <- do.call(dplyr::bind_rows, args = dfs)
  df <- df[!duplicated(df$short_path), ]
  df
}

#' Get all snippets from all sources where snippets can exist
#'
#' Priority is given to the current project, then to git snippets,
#' then imported projects, then builtin snippets.
#'
#' @return A dataframe where each row represents a snippet two columns:
#'   - full_path: The full normalized path to the snippet
#'   - short_path: A path to the snippet starting at the input path
merge_all_snippets <- function() {
  paths <- c(
    # The order of this list matters; if a file appears in the same path in multiple
    # sources, the latest one will take precedence
    get_snippets_paths_current_project(),
    get_snippets_paths_git(),
    get_snippets_paths_imported_projects(),
    get_snippets_paths_builtin()
  )
  find_snippets(paths)
}

#' Get paths of all writable directories where snippets can exist
#'
#' Writable directories are considered to be the current project snippets
#' directory, and any git repositories that the user has write permission to.
#'
#' @return A named list with 2 values:
#'   - Current Project: path to the snippets directory in the current project
#'   - Git Repos: named list of paths to writable git repositories
get_editable_snippets_paths <- function() {
  git_paths <- get_writable_git_repos()
  paths <- list("Current Project" = get_user_project_dir())
  if (length(git_paths) > 0) {
    paths[["Git Repos"]] <- stats::setNames(git_paths, basename(git_paths))
  }
  paths
}

#' Add a snippet
#'
#' @param contents The code in the snippet
#' @param name The name of the snippet file
#' @param repo The full path to the project that hosts the library of snippets
#' @param local_folder The local path to the snippet file
#' @return The full path to the newly created snippet
add_snippet <- function(contents, name, repo, local_folder) {
  snippet_file_name <- paste0(name, ".R")
  snippet_dir <- file.path(repo, "snippets", local_folder)
  if (!dir.exists(snippet_dir)) {
    dir.create(snippet_dir, showWarnings = FALSE, recursive = TRUE)
  }
  snippet_path <- file.path(snippet_dir, snippet_file_name)
  writeLines(contents, snippet_path)
  snippet_path
}

#' Edit a snippet
#'
#' @param contents The new code in the snippet
#' @param file The full path to the snippet file
edit_snippet <- function(contents, file) {
  writeLines(contents, file)
}
