#' A subclass of Transformation for group+aggregate operations
#' @field cols Columns to group by
#' @field aggregations Named list of aggregations, where the names are the column
#' names and the values are the aggregate function
AggregateTransformation <- R6::R6Class(
  "AggregateTransformation",
  inherit = Transformation,

  private = list(

    .cols = NULL,
    .aggregations = NULL,

    get_dependencies = function() {
      if (private$.tidyverse) "dplyr" else NULL
    },

    get_full_code = function(name_in) {
      aggregations <- private$get_code_for_aggregators(name_in)
      if (self$tidyverse) {
        if (length(aggregations) == 1) {
          summaries <- glue::glue("  summarise({aggregations[[1]]})")
        } else {
          summaries <- glue::glue(
            '  summarise(\n',
            '{paste("    ", aggregations, collapse = ",\n")}\n',
            '  )',
            .trim = FALSE
          )
        }
        glue::glue(
          '{name_in} %>%\n  group_by({paste(private$.cols, collapse = ", ")}) %>%\n',
          summaries,
          ' %>%\n  ungroup()',
          .trim = FALSE
        )
      } else {
        if (length(aggregations) == 1) {
          glue::glue(
            '{aggregations[[1]]}'
          )
        } else {
          glue::glue(
            'Reduce(',
            '  merge,',
            '  list(',
            glue::glue_collapse(paste0("    ", aggregations), sep = ",\n"),
            '  )',
            ')',
            .sep = '\n'
          )
        }
      }
    },

    get_code_for_aggregators = function(name_in) {
      if (self$tidyverse) {
        aggregations <- lapply(seq_along(self$aggregations), function(idx) {
          type <- self$aggregations[[idx]]
          col <- names(self$aggregations[idx])
          if (! type %in% AggregateTransformation$OPTIONS) {
            stop("The aggregator must be one of: ", paste(AggregateTransformation$OPTIONS, collapse = " "), " (given: ", type, ")", call. = FALSE)
          }
          private$get_code_for_aggregator(type, col, name_in)
        })
      } else {
        aggr_types <- unique(self$aggregations)
        aggregations <- lapply(aggr_types, function(type) {
          if (! type %in% AggregateTransformation$OPTIONS) {
            stop("The aggregator must be one of: ", paste(AggregateTransformation$OPTIONS, collapse = " "), " (given: ", type, ")", call. = FALSE)
          }
          aggr_cols <- names(self$aggregations[self$aggregations == type])
          aggr_cols <- unique(aggr_cols)
          private$get_code_for_aggregator(type, aggr_cols, name_in)
        })
      }
    },

    get_code_for_aggregator = function(type, aggr_cols, name_in) {
      if (type == "size") {
        fxn <- "length"
      } else {
        fxn <- type
      }

      if (self$tidyverse) {
        glue::glue(
          "{aggr_cols}_{type} = {fxn}({aggr_cols})"
        )
      } else {
        glue::glue(
          "aggregate(",
          "cbind(",
          glue::glue_collapse(paste(paste(aggr_cols, type, sep = '_'), aggr_cols, sep = ' = '), sep = ', '),
          ")",
          " ~ ",
          "{glue::glue_collapse(self$cols, sep = ' + ')}, ",
          "{name_in}, ",
          "FUN = {fxn})"
        )
      }
    }

  ),

  active = list(
    cols = function() {
      private$.cols
    },
    aggregations = function() {
      private$.aggregations
    }
  ),

  public = list(

    #' @description
    #' Create a new instance
    #'
    #' @param cols Columns to group by
    #' @param aggregations Named list of aggregations, where the names are the column
    #' names and the values are the aggregate function
    #' @param name_out The name of the resulting dataframe
    #' @param tidyverse Whether this transformation should use tidyverse code or base R
    initialize = function(cols, aggregations, name_out = NULL, tidyverse = NULL) {
      if (length(cols) == 0) {
        stop("You must provide at least one column to group by", call. = FALSE)
      }
      if (length(aggregations) == 0) {
        stop("You must provide at least one aggregation", call. = FALSE)
      }
      super$initialize(name_out, tidyverse)
      private$.cols <- cols
      private$.aggregations <- aggregations
      invisible(self)
    },

    #' @description
    #' Print a string representation
    print = function() {
      super$print()
      cat0(
        "\n",
        "    Group by: ",
        paste(self$cols, collapse = ", "), "\n",
        "    Aggregate by: ",
        paste(names(self$aggregations), unname(self$aggregations), sep = "=", collapse = ", "), "\n"
      )
    }
  )
)

AggregateTransformation$OPTIONS <- list(
  SUM    = "sum",
  MEAN   = "mean",
  MIN    = "min",
  MAX    = "max",
  MEDIAN = "median",
  SD     = "sd",
  LENGTH = "size"
)

AggregateTransformation$name_short <- "aggregate"
AggregateTransformation$name_long <- "Group & aggregate"

#' Shiny module: Create a AggregateTransformation
#'
#' @param id Unique ID for the module
#' @param data (reactive) A data frame to select columns from
#' @param old_xform (reactive) A transformation to pre-populate the fields, or `NULL`
#' @return List of reactives:
#'   - validate: whether or not the current selections are valid
#'   - xform: The resulting Transformation
#' @name AggregateTransformationShiny
AggregateTransformation$shiny <- list(

  ui = function(id) {
    ns <- NS(id)

    tagList(
      shinyjs::useShinyjs(),
      fluidRow(
        column(3, shinyWidgets::pickerInput(ns("aggregate_cols"), "Group by", NULL, multiple = TRUE)),
        column(6, pairs_selector_ui(ns("aggregate_pairs"), "Aggregate on", "Aggregator", choices_2 = as.character(AggregateTransformation$OPTIONS))),
        column(3, textInput(ns("aggregate_name"), "New name", "dfg"))
      )
    )
  },

  server = function(id, data = reactive(NULL), old_xform = reactive(NULL)) {
    moduleServer(
      id,
      function(input, output, session) {

        ns <- session$ns

        aggregate_pairs <- pairs_selector_server("aggregate_pairs", choices_df_1 = data)

        observeEvent(data(), {
          shinyWidgets::updatePickerInput(session, "aggregate_cols", choices = names(data()))
        })

        observeEvent(old_xform(), {
          if (inherits(old_xform(), AggregateTransformation$classname)) {
            shinyWidgets::updatePickerInput(session, "xform_type", selected = "aggregate")
            shinyWidgets::updatePickerInput(session, "aggregate_cols", selected = old_xform()$cols)
            updateTextInput(session, "aggregate_name", value = old_xform()$name_out)
            aggregate_pairs$set(old_xform()$aggregations)
          }
        })

        observeEvent(input$aggregate_name, {
          if (!is_valid_name(input$aggregate_name)) {
            updateTextInput(session, "aggregate_name", value = make.names(input$aggregate_name))
          }
        })

        validate <- reactive({
          length(input$aggregate_cols) > 0 && is_valid_name(input$aggregate_name) &&
            length(aggregate_pairs$get()) > 0
        })

        xform <- reactive({
          AggregateTransformation$new(cols = input$aggregate_cols, aggregations = aggregate_pairs$get(), name_out = input$aggregate_name)
        })

        return(list(
          validate = validate,
          xform = xform
        ))

      }
    )
  }
)
