#' A subclass of Transformation for dropping columns
#' @field cols Columns to remove (at least one)
DropTransformation <- R6::R6Class(
  "DropTransformation",
  inherit = Transformation,

  private = list(
    .cols = NULL,

    get_dependencies = function() {
      if (private$.tidyverse) "dplyr" else NULL
    },

    get_full_code = function(name_in) {
      if (private$.tidyverse) {
        glue::glue(
          '{name_in} %>% ',
          'select(-c("{paste(self$cols, collapse = \'", "\')}"))'
        )
      } else {
        glue::glue(
          '{name_in}',
          '[, !(names({name_in}) %in% c("{paste(self$cols, collapse = \'", "\')}")), drop = FALSE]'
        )
      }
    }
  ),

  active = list(
    cols = function() {
      private$.cols
    }
  ),

  public = list(

    #' @description
    #' Create a new instance
    #'
    #' @param cols Columns to remove (at least one)
    #' @param name_out The name of the resulting dataframe
    #' @param tidyverse Whether this transformation should use tidyverse code or base R
    initialize = function(cols, name_out = NULL, tidyverse = NULL) {
      if (length(cols) == 0) {
        stop("You must provide at least one column", call. = FALSE)
      }
      super$initialize(name_out, tidyverse)
      private$.cols <- cols
      invisible(self)
    },

    #' @description
    #' Print a string representation
    print = function() {
      super$print()
      cat0("Columns: ", paste(self$cols, collapse = ", "), "\n")
    }
  )
)

DropTransformation$name_short <- "drop"
DropTransformation$name_long <- "Remove columns"

#' Shiny module: Create a DropTransformation
#'
#' @param id Unique ID for the module
#' @param data (reactive) A data frame to remove columns from
#' @param old_xform (reactive) A transformation to pre-populate the fields, or `NULL`
#' @return List of reactives:
#'   - validate: whether or not the current selections are valid
#'   - xform: The resulting Transformation
#' @name DropTransformationShiny
DropTransformation$shiny <- list(

  ui = function(id) {
    ns <- NS(id)

    fluidRow(
      column(4, shinyWidgets::pickerInput(ns("drop_cols"), "Columns", NULL, multiple = TRUE)),
      column(2, textInput(ns("drop_name"), "New name", "df"))
    )
  },

  server = function(id, data = reactive(NULL), old_xform = reactive(NULL)) {
    moduleServer(
      id,
      function(input, output, session) {

        observeEvent(data(), {
          shinyWidgets::updatePickerInput(session, "drop_cols", choices = names(data()))
        })

        observeEvent(old_xform(), {
          if (inherits(old_xform(), DropTransformation$classname)) {
            shinyWidgets::updatePickerInput(session, "drop_cols", selected = old_xform()$cols)
            updateTextInput(session, "drop_name", value = old_xform()$name_out)
          }
        })

        observeEvent(input$drop_name, {
          if (!is_valid_name(input$drop_name)) {
            updateTextInput(session, "drop_name", value = make.names(input$drop_name))
          }
        })

        validate <- reactive({
          length(input$drop_cols) > 0 && is_valid_name(input$drop_name)
        })

        xform <- reactive({
          DropTransformation$new(cols = input$drop_cols, name_out = input$drop_name)
        })

        return(list(
          validate = validate,
          xform = xform
        ))

      }
    )
  }
)
