#' A subclass of Transformation for filter operations
#' @field col The column to filter
#' @field op The filter operation (such as equals, less than, greater or equal to, etc.)
#' @field value The value to filter on
FilterTransformation <- R6::R6Class(
  "FilterTransformation",
  inherit = Transformation,

  private = list(
    .col = NULL,
    .op = NULL,
    .value = NULL,
    .type = NULL,

    get_dependencies = function() {
      if (private$.tidyverse) "dplyr" else NULL
    },

    get_full_code = function(name_in) {
      value <- self$value
      if (!is.null(private$.type) && private$.type %in% c("character", "factor")) {
        value <- glue::glue('{shQuote(value, type = "cmd")}')
      }

      if (private$.tidyverse) {
        glue::glue(
          '{name_in} %>% ',
          'filter({self$col} {self$op} {value})'
        )
      } else {
        glue::glue(
          '{name_in}',
          '[ {name_in}[["{self$col}"]] {self$op} {value}, , drop = FALSE]'
        )
      }
    }
  ),

  active = list(
    col = function() {
      private$.col
    },
    op = function() {
      private$.op
    },
    value = function() {
      private$.value
    }
  ),

  public = list(

    #' @description
    #' Create a new instance
    #'
    #' @param col The column to filter
    #' @param op The filter operation (such as equals, less than, greater or equal to, etc.)
    #' @param value The value to filter on
    #' @param type The class of the filter value, to determine whether or not quotes
    #' should be placed around the value if it's string-like
    #' @param name_out The name of the resulting dataframe
    #' @param tidyverse Whether this transformation should use tidyverse code or base R
    initialize = function(col, op, value, type = NULL, name_out = NULL, tidyverse = NULL) {
      if (length(col) != 1) {
        stop("You must provide exactly one column", call. = FALSE)
      }
      if (! op %in% FilterTransformation$OPTIONS) {
        stop("The operation must be one of: ", paste(FilterTransformation$OPTIONS, collapse = " "), call. = FALSE)
      }
      super$initialize(name_out, tidyverse)
      private$.col <- col
      private$.op <- op
      private$.value <- value
      private$.type <- type
      invisible(self)
    },

    #' @description
    #' Print a string representation
    print = function() {
      super$print()
      cat0(glue::glue("{self$col} {self$op} {self$value}",
                      if (is.null(private$.type)) "" else " ({private$.type})"), "\n")
    }
  )
)

FilterTransformation$OPTIONS <- list(
  LESS_THAN    = "<",
  GREATER_THAN = ">",
  LESS_EQ      = "<=",
  GREATER_EQ   = ">=",
  EQUALS       = "==",
  NOT_EQUALS   = "!="
)

FilterTransformation$name_short <- "filter"
FilterTransformation$name_long <- "Filter rows"

#' Shiny module: Create a FilterTransformation
#'
#' @param id Unique ID for the module
#' @param data (reactive) A data frame to select columns from
#' @param old_xform (reactive) A transformation to pre-populate the fields, or `NULL`
#' @return List of reactives:
#'   - validate: whether or not the current selections are valid
#'   - xform: The resulting Transformation
#' @name FilterTransformationShiny
FilterTransformation$shiny <- list(

  ui = function(id) {
    ns <- NS(id)

    fluidRow(
      column(4, shinyWidgets::pickerInput(ns("filter_col"), "Column", NULL)),
      column(2, shinyWidgets::pickerInput(ns("filter_op"), "Operation", unname(FilterTransformation$OPTIONS))),
      column(2, textInput(ns("filter_value"), "Value", "")),
      column(2, textInput(ns("filter_name"), "New name", "df"))
    )
  },

  server = function(id, data = reactive(NULL), old_xform = reactive(NULL)) {
    moduleServer(
      id,
      function(input, output, session) {

        observeEvent(data(), {
          shinyWidgets::updatePickerInput(session, "filter_col", choices = names(data()))
        })

        observeEvent(old_xform(), {
          if (inherits(old_xform(), FilterTransformation$classname)) {
            shinyWidgets::updatePickerInput(session, "filter_col", selected = old_xform()$col)
            shinyWidgets::updatePickerInput(session, "filter_op", selected = old_xform()$op)
            updateTextInput(session, "filter_value", value = old_xform()$value)
            updateTextInput(session, "filter_name", value = old_xform()$name_out)
          }
        })

        observeEvent(input$filter_name, {
          if (!is_valid_name(input$filter_name)) {
            updateTextInput(session, "filter_name", value = make.names(input$filter_name))
          }
        })

        validate <- reactive({
          is_valid_name(input$filter_name)
        })

        xform <- reactive({
          req(input$filter_col)
          col_type <- class(data()[[input$filter_col]])
          col_type <- utils::tail(col_type, 1)
          FilterTransformation$new(col = input$filter_col, op = input$filter_op, value = input$filter_value, type = col_type, name_out = input$filter_name)
        })

        return(list(
          validate = validate,
          xform = xform
        ))

      }
    )
  }
)
