#' A subclass of Transformation for merge operations
#' @field df The name of the dataframe to merge
#' @field type The type of merge to perform
#' @field by Named list of the column names to use for the merge, or `NULL`
#' to do natural merging (merge by all common column names)
MergeTransformation <- R6::R6Class(
  "MergeTransformation",
  inherit = Transformation,

  private = list(
    .df = NULL,
    .type = NULL,
    .by = NULL,

    get_dependencies = function() {
      if (private$.tidyverse) "dplyr" else NULL
    },

    get_full_code = function(name_in) {
      by <- self$by
      same_cols <- which(names(by) == unname(by))
      same_cols <- unname(by)[same_cols]
      diff_cols <- which(names(by) != unname(by))
      diff_cols <- by[diff_cols]

      if (private$.tidyverse) {
        if (self$type == MergeTransformation$OPTIONS$INNER_JOIN) {
          fxn <- "inner_join"
        } else if (self$type == MergeTransformation$OPTIONS$OUTER_JOIN) {
          fxn <- "full_join"
        } else if (self$type == MergeTransformation$OPTIONS$LEFT_JOIN) {
          fxn <- "left_join"
        } else if (self$type == MergeTransformation$OPTIONS$RIGHT_JOIN) {
          fxn <- "right_join"
        }

        if (is.null(by)) {
          by_str <- ""
        } else {
          same_cols_str <- glue::glue('"{ paste(same_cols, collapse = \'", "\') }"')
          diff_cols_str <- glue::glue('"{ paste(names(diff_cols), unname(diff_cols), sep = \'" = "\', collapse = \'", "\') }"')
          if (length(diff_cols) == 0) {
            by_str <- same_cols_str
          } else if (length(same_cols) == 0) {
            by_str <- diff_cols_str
          } else {
            by_str <- paste0(same_cols_str, ", ", diff_cols_str)
          }

          if (length(same_cols) == 1 && length(diff_cols) == 0) {
            by_str <- glue::glue(", by = { by_str }")
          } else {
            by_str <- glue::glue(", by = c({ by_str })")
          }
        }

        glue::glue(
          '{name_in} %>% ',
          '{ fxn }({ self$df }{ by_str })'
        )
      } else {
        if (self$type == MergeTransformation$OPTIONS$INNER_JOIN) {
          type_str <- ""
        } else if (self$type == MergeTransformation$OPTIONS$OUTER_JOIN) {
          type_str <- ", all = TRUE"
        } else if (self$type == MergeTransformation$OPTIONS$LEFT_JOIN) {
          type_str <- ", all.x = TRUE"
        } else if (self$type == MergeTransformation$OPTIONS$RIGHT_JOIN) {
          type_str <- ", all.y = TRUE"
        }

        if (is.null(by)) {
          by_str <- ""
        } else {
          same_cols_str <- glue::glue('"{ paste(same_cols, collapse = \'", "\') }"')
          diff_cols_str_x <- glue::glue('"{ paste(names(diff_cols), collapse = \'", "\') }"')
          diff_cols_str_y <- glue::glue('"{ paste(unname(diff_cols), collapse = \'", "\') }"')
          if (length(diff_cols) == 0) {
            if (length(same_cols) == 1) {
              by_str <- glue::glue('by = { same_cols_str }')
            } else {
              by_str <- glue::glue('by = c({ same_cols_str })')
            }
          } else if (length(same_cols) == 0) {
            if (length(diff_cols) == 1) {
              by_str <- glue::glue(
                'by.x = { diff_cols_str_x }, ',
                'by.y = { diff_cols_str_y }'
              )
            } else {
              by_str <- glue::glue(
                'by.x = c({ diff_cols_str_x }), ',
                'by.y = c({ diff_cols_str_y })'
              )
            }
          } else {
            by_str <- glue::glue(
              'by.x = c({ same_cols_str }, { diff_cols_str_x }), ',
              'by.y = c({ same_cols_str }, { diff_cols_str_y })',
            )
          }

          by_str <- glue::glue(", { by_str }")
        }

        glue::glue(
          'merge({ name_in }, { self$df }{ by_str}{ type_str })',
        )
      }
    }
  ),

  active = list(
    df = function() {
      private$.df
    },
    type = function() {
      private$.type
    },
    by = function() {
      private$.by
    }
  ),

  public = list(

    #' @description
    #' Create a new instance
    #'
    #' @param df The name of the dataframe to merge
    #' @param type The type of merge to perform (inner, outer, left, right)
    #' @param by Named list of the column names to use for the merge, or `NULL` to do
    #' natural merging (merge by all common column names). If a named list is provided,
    #' the names are the left dataframe column names and the values are the right dataframe
    #' column names.
    #' @param name_out The name of the resulting dataframe
    #' @param tidyverse Whether this transformation should use tidyverse code or base R
    initialize = function(df, type, by = NULL, name_out = NULL, tidyverse = NULL) {
      if (!is.null(by) && (is.null(names(by)) || ("" %in% names(by)))) {
        stop("You must provide a named list to merge the dataframes by, or `NULL`", call. = FALSE)
      }
      if (! type %in% MergeTransformation$OPTIONS) {
        stop("The type must be one of: ", paste(MergeTransformation$OPTIONS, collapse = " "), call. = FALSE)
      }
      super$initialize(name_out, tidyverse)
      private$.df <- df
      private$.type <- type
      private$.by <- by
      invisible(self)
    },

    #' @description
    #' Print a string representation
    print = function() {
      super$print()

      by <- self$by
      if (is.null(by)) {
        columns_str <- "natural merging"
      } else {
        same_cols <- which(names(by) == unname(by))
        same_cols <- unname(by)[same_cols]
        diff_cols <- which(names(by) != unname(by))
        diff_cols <- by[diff_cols]
        columns_str <- ""
        if (length(diff_cols) == 0) {
          columns_str <- paste(same_cols, collapse = ", ")
        } else {
          if (length(same_cols) > 0) {
            columns_str <- paste0(paste(same_cols, collapse = ", "), ", ")
          }
          diff_str <- paste(names(diff_cols), unname(diff_cols), sep = "=", collapse = ", ")
          columns_str <- paste0(columns_str, diff_str)
        }
        columns_str <- paste0("columns: ", columns_str)
      }

      cat0(glue::glue("{ self$type } merge with { self$df } using {columns_str}"), "\n")
    }
  )
)

MergeTransformation$OPTIONS <- list(
  INNER_JOIN = "inner",
  LEFT_JOIN = "left",
  RIGHT_JOIN = "right",
  OUTER_JOIN = "outer"
)

MergeTransformation$name_short <- "merge"
MergeTransformation$name_long <- "Merge dataframes"

#' Shiny module: Create a MergeTransformation
#'
#' @param id Unique ID for the module
#' @param data (reactive) A data frame to select columns from
#' @param old_xform (reactive) A transformation to pre-populate the fields, or `NULL`
#' @return List of reactives:
#'   - validate: whether or not the current selections are valid
#'   - xform: The resulting Transformation
#' @name MergeTransformationShiny
MergeTransformation$shiny <- list(

  ui = function(id) {
    ns <- NS(id)

    tagList(
      shinyjs::useShinyjs(),
      fluidRow(
        column(3, shinyWidgets::pickerInput(ns("df"), "Dataframe to merge", NULL, multiple = TRUE, options = list(`max-options` = 1))),
        column(3, textInput(ns("name"), "New name", "df_merged"))
      ),
      br(),
      shinyWidgets::radioGroupButtons(
        ns("type"), NULL,
        choiceNames = firstup(as.character(MergeTransformation$OPTIONS)),
        choiceValues = as.character(MergeTransformation$OPTIONS)
      ),
      br(),
      shinyWidgets::materialSwitch(
        ns("natural"), "Join on all common clumns", status = "primary", value = TRUE, right = TRUE
      ),
      conditionalPanel(
        "input.natural", ns = ns,
        textOutput(ns("joining_on_natural"))
      ),
      conditionalPanel(
        "!input.natural", ns = ns,
        pairs_selector_ui(ns("merge_pairs"), "Merge by (x)", "Merge by (y)")
      )
    )
  },

  server = function(id, data = reactive(NULL), old_xform = reactive(NULL)) {
    moduleServer(
      id,
      function(input, output, session) {

        ns <- session$ns

        observeEvent(data(), {
          shinyWidgets::updatePickerInput(session, "df", choices = get_common_columns_all_dfs(data()))
        })

        df <- reactive({
          if (is.null(input$df)) {
            NULL
          } else {
            get(input$df, envir = .GlobalEnv)
          }
        })

        merge_pairs <- pairs_selector_server("merge_pairs", choices_df_1 = data, choices_df_2 = df)

        observeEvent(df(), {
          cols <- intersect(names(data()), names(df()))
          if (length(cols) == 0) {
            shinyWidgets::updateMaterialSwitch(session, "natural", value = FALSE)
            shinyjs::disable("natural")
          } else {
            shinyjs::enable("natural")
          }
        })

        observeEvent(old_xform(), {
          if (inherits(old_xform(), MergeTransformation$classname)) {
            shinyWidgets::updatePickerInput(session, "xform_type", selected = "merge")
            shinyWidgets::updatePickerInput(session, "df", selected = old_xform()$df)
            shinyWidgets::updateRadioGroupButtons(session, "type", selected = old_xform()$type)
            updateTextInput(session, "name", value = old_xform()$name_out)

            natural <- is.null(old_xform()$by)
            shinyWidgets::updateMaterialSwitch(session, "natural", value = natural)
            merge_pairs$set(old_xform()$by)
          }
        })

        observeEvent(input$name, {
          if (!is_valid_name(input$name)) {
            updateTextInput(session, "name", value = make.names(input$name))
          }
        })

        validate <- reactive({
          !is.null(input$df) && is_valid_name(input$name) &&
            (input$natural || length(merge_pairs$get()) > 0)
        })

        xform <- reactive({
          if (input$natural) {
            by <- NULL
          } else {
            by <- merge_pairs$get()
          }
          MergeTransformation$new(df = input$df, type = input$type, by = by, name_out = input$name)
        })

        output$joining_on_natural <- renderText({
          req(df())
          df1 <- data()
          df2 <- df()
          cols <- intersect(names(df1), names(df2))
          paste("Join on columns:", paste(cols, collapse = ", "))
        })

        return(list(
          validate = validate,
          xform = xform
        ))
      }
    )
  }
)

get_common_columns_all_dfs <- function(df) {
  df_names <- sapply(get_dataframes(), USE.NAMES = TRUE, simplify = FALSE, function(other_df) {
    cols <- intersect(names(df), names(get(other_df, envir = .GlobalEnv)))
    if (length(cols) == 0) {
      common <- "(no common columns)"
    } else {
      common <- paste0("(", length(cols), " common columns)")
    }
    paste(other_df, common)
  })
  if (length(df_names) > 0) {
    df_names <- stats::setNames(names(df_names), df_names)
  }
  df_names
}
