#' A subclass of Transformation for select operations
#' @field cols Columns to keep (at least one)
SelectTransformation <- R6::R6Class(
  "SelectTransformation",
  inherit = Transformation,

  private = list(
    .cols = NULL,

    get_dependencies = function() {
      if (private$.tidyverse) "dplyr" else NULL
    },

    get_full_code = function(name_in) {
      if (private$.tidyverse) {
        glue::glue(
          '{name_in} %>% ',
          'select(c("{paste(self$cols, collapse = \'", "\')}"))'
        )
      } else {
        glue::glue(
          '{name_in}',
          '[c("{paste(self$cols, collapse = \'", "\')}")]'
        )
      }
    }
  ),

  active = list(
    cols = function() {
      private$.cols
    }
  ),

  public = list(

    #' @description
    #' Create a new instance
    #'
    #' @param cols Columns to keep (at least one)
    #' @param name_out The name of the resulting dataframe
    #' @param tidyverse Whether this transformation should use tidyverse code or base R
    initialize = function(cols, name_out = NULL, tidyverse = NULL) {
      if (length(cols) == 0) {
        stop("You must provide at least one column", call. = FALSE)
      }
      super$initialize(name_out, tidyverse)
      private$.cols <- cols
      invisible(self)
    },

    #' @description
    #' Print a string representation
    print = function() {
      super$print()
      cat0("Columns: ", paste(self$cols, collapse = ", "), "\n")
    }
  )
)

SelectTransformation$name_short <- "select"
SelectTransformation$name_long <- "Select columns"

#' Shiny module: Create a SelectTransformation
#'
#' @param id Unique ID for the module
#' @param data (reactive) A data frame to select columns from
#' @param old_xform (reactive) A transformation to pre-populate the fields, or `NULL`
#' @return List of reactives:
#'   - validate: whether or not the current selections are valid
#'   - xform: The resulting Transformation
#' @name SelectTransformationShiny
SelectTransformation$shiny <- list(

  ui = function(id) {
    ns <- NS(id)

    fluidRow(
      column(4, shinyWidgets::pickerInput(ns("select_cols"), "Columns", NULL, multiple = TRUE)),
      column(2, textInput(ns("select_name"), "New name", "df"))
    )
  },

  server = function(id, data = reactive(NULL), old_xform = reactive(NULL)) {
    moduleServer(
      id,
      function(input, output, session) {

        observeEvent(data(), {
          shinyWidgets::updatePickerInput(session, "select_cols", choices = names(data()))
        })

        observeEvent(old_xform(), {
          if (inherits(old_xform(), SelectTransformation$classname)) {
            shinyWidgets::updatePickerInput(session, "select_cols", selected = old_xform()$cols)
            updateTextInput(session, "select_name", value = old_xform()$name_out)
          }
        })

        observeEvent(input$select_name, {
          if (!is_valid_name(input$select_name)) {
            updateTextInput(session, "select_name", value = make.names(input$select_name))
          }
        })

        validate <- reactive({
          length(input$select_cols) > 0 && is_valid_name(input$select_name)
        })

        xform <- reactive({
          SelectTransformation$new(cols = input$select_cols, name_out = input$select_name)
        })

        return(list(
          validate = validate,
          xform = xform
        ))

      }
    )
  }
)
