#' R6 superclass for storing information about a transformation
#'
#' @description
#' A subclass must implement:
#'   - `private$get_full_code(name_in)`
#'   - `private$get_dependencies()`
#' @field name_out The name of the resulting dataframe
#' @field tidyverse Whether this transformation should use tidyverse code or base R
#' @field dependencies List of dependency packages used by the code in this transformation
Transformation <- R6::R6Class(
  "Transformation",
  cloneable = FALSE,

  private = list(
    .name_out = NULL,
    .tidyverse = NULL
  ),

  active = list(
    name_out = function() {
      private$.name_out
    },

    tidyverse = function() {
      private$.tidyverse
    },

    dependencies = function() {
      private$get_dependencies()
    }
  ),

  public = list(

    #' @description
    #' Create a new instance
    #'
    #' @param name_out The name of the resulting dataframe
    #' @param tidyverse Whether this transformation should use tidyverse code or base R
    initialize = function(name_out = NULL, tidyverse = NULL) {
      private$.name_out <- name_out %||% "df"
      self$use_tidyverse(tidyverse %||% FALSE)
      invisible(self)
    },

    #' @description
    #' Print a string representation
    print = function() {
      cat0(
        "<", class(self)[1], "> (", self$name_out, " | ",
        if (private$.tidyverse) "{tidyverse}" else "base R",
        ") "
      )
    },

    #' @description
    #' Whether this transformation should use tidyverse code or base R
    #'
    #' @param use A boolean whether to use tidyverse or not
    use_tidyverse = function(use) {
      if (!is.null(use)) {
        private$.tidyverse <- use
      }
      invisible(self)
    },

    #' @description
    #' Get the code of the transformation as a single string
    #'
    #' @param name_in The name of the input dataframe
    #' @param dependencies Whether or not to include `library()` calls
    get_code = function(name_in, dependencies = TRUE) {
      code <- paste(self$name_out, "<-", private$get_full_code(name_in))
      if (dependencies && length(private$get_dependencies()) >= 1) {
        code <- paste0(
          paste(paste0("library(", private$get_dependencies(), ")\n"), collapse = ""),
          code
        )
      }
      code
    }
  )
)

ALL_XFORMS <- c(
  SelectTransformation,
  DropTransformation,
  FilterTransformation,
  MissingValuesTransformation,
  AggregateTransformation,
  MergeTransformation
)
