#' Get the Domino user's project directory
get_user_project_dir <- function() {
  if (is_project_git_based()) {
    Sys.getenv("DOMINO_ARTIFACTS_DIR", getwd())
  } else {
    Sys.getenv("DOMINO_WORKING_DIR", getwd())
  }
}

#' Get the Domino user's datasets directory
get_user_datasets_dir <- function() {
  Sys.getenv("DOMINO_DATASETS_DIR", getwd())
}

#' Get the Domino user's git repositories root directory
get_user_git_dir <- function() {
  Sys.getenv("DOMINO_IMPORTED_CODE_DIR", "")
}

#' Get a list of the Domino user's git repositories
get_user_git_repos <- function() {
  dir <- get_user_git_dir()
  if (dir == "") {
    character(0)
  } else {
    list.dirs(dir, recursive = FALSE)
  }
}

#' Get a list of directories of the Domino user's imported projects
get_imported_projects_dirs <- function() {
  project_vars <- grep("^DOMINO_.*_WORKING_DIR$", names(Sys.getenv()), value = TRUE)
  if (length(project_vars) == 0) {
    character(0)
  } else {
    as.character(Sys.getenv(project_vars))
  }
}

#' Get the current Domino project name
get_project_name <- function() {
  Sys.getenv("DOMINO_PROJECT_NAME", "sample_project")
}

#' Is the current project a git-based project?
is_project_git_based <- function() {
  Sys.getenv("DOMINO_IS_GIT_BASED") == "true"
}

#' Get the path where user-uploaded files should be saved
get_user_upload_dir <- function() {
  if (is_project_git_based()) {
    file.path(
      get_user_datasets_dir(),
      get_project_name()
    )
  } else {
    file.path(
      get_user_datasets_dir(),
      "local",
      get_project_name()
    )
  }
}

#' Get the Domino API key
get_api_key <- function() {
  key <- Sys.getenv("DOMINO_USER_API_KEY", "")
  if (key == "") {
    stop("Could not find envvar 'DOMINO_USER_API_KEY'", call. = FALSE)
  }
  key
}

#' Get the Domino API base URL
get_api_base <- function() {
  url <- Sys.getenv("DOMINO_API_HOST", "")
  if (url == "") {
    stop("Could not find envvar 'DOMINO_API_HOST'", call. = FALSE)
  }
  url
}

#' Get the Domino platform version
get_domino_version <- function() {
  version <- tryCatch({
    resp <- call_api("/version")
    resp$version
  }, error = function(err) {
    "0"
  })
  version
}

#' Get the Domino user ID
get_user_id <- function() {
  id <- tryCatch({
    resp <- call_api("/v4/users/self")
    resp$id
  }, error = function(err) {
    "0"
  })
  id
}

#' Get the Domino customer's setting on allowing Mixpanel via client-side
get_mixpanel_enabled_client <- function() {
  enabled <- tryCatch({
    resp <- call_api("/v4/auth/principal")
    isTRUE(resp$mixpanelSettings$frontendClientEnabled)
  }, error = function(err) {
    FALSE
  })
  enabled
}

#' Get the Domino customer's setting on allowing Mixpanel via server-side
get_mixpanel_enabled_server <- function() {
  enabled <- tryCatch({
    resp <- call_api("/v4/auth/principal")
    isTRUE(resp$mixpanelSettings$backendClientEnabled)
  }, error = function(err) {
    FALSE
  })
  enabled
}

#' Get a list of git writable repositories for the current Domino user
get_writable_git_repos <- function() {
  repos <- get_user_git_repos()
  writable <- unlist(lapply(repos, is_git_writable))
  if (length(writable) == 0) {
    return(character(0))
  }
  writable_idx <- which(writable)
  repos[writable_idx]
}

#' Get a list of all DominoDataSource objects for the current Domino user
get_data_sources <- function() {
  ds <- tryCatch({
    sources <- get_data_sources_raw()
    lapply(sources, function(source) {
      DominoDataSource$new(id = source$id, name = source$name, type = get_data_source_shortname(source$dataSourceType), config = source$config)
    })
  }, error = function(err) {
    list()
  })
  ds
}

#' Get a raw list of the Domino Data Sources for the current Domino user
get_data_sources_raw <- function() {
  tryCatch({
    endpoint <- glue::glue("/v4/datasource/projects/{ Sys.getenv('DOMINO_PROJECT_ID', '') }?authenticatedOnly=true")
    call_api(endpoint)
  }, error = function(err) {
    list()
  })
}

#' Get whether or not the current Domino deployment supports Data Sources
is_data_sources_supported <- function() {
  is_data_sources_enabled() &&
    as.numeric_version(get_domino_version()) >= as.numeric_version(DATASOURCES_MIN_VERSION)
}

#' Whether or not Data Sources are enabled
is_data_sources_enabled <- function() {
  Sys.getenv('DCA_R_DATASOURCES_DISABLED', "") == ""
}

#' Whether or not Snippets editing is enabled
is_snippet_editing_enabled <- function() {
  Sys.getenv('DCA_R_SNIPPET_EDITING_DISABLED', "") == ""
}
