#' Run a `Transformation` on a dataframe
#'
#' @param data The dataframe to run on
#' @param xform The transformation
#' @param quiet Whether or not to suppress messages and warnings
#' @return The dataframe result from running the transformation
run_xform <- function(data, xform, quiet = FALSE) {
  if (quiet) {
    suppressWarnings(suppressMessages(eval(parse(text = xform$get_code("data")))))
  } else {
    eval(parse(text = xform$get_code("data")))
  }
}

#' Clean the search path by detaching any packages not available earlier
#'
#' @param init The list of packages to keep
clean_search_path <- function(init) {
  sapply(setdiff(search(), init), function(name) detach(name, character.only = TRUE))
}

#' Determine the tidyverse status of each transformation in a sequence of transformations
#'
#' @param xformseq A `TransformationSequence`
#' @return A boolean vector with the length of the number of transformations in the input
get_tidyverse_status <- function(xformseq) {
  sapply(xformseq$transformations, function(xform) xform$tidyverse)
}

get_xform_sel1 <- function() {
  SelectTransformation$new("col1")
}
get_xform_sel2 <- function() {
  SelectTransformation$new(c("col1", "col2"))
}
get_xform_drop1 <- function() {
  DropTransformation$new("col1")
}
get_xform_drop2 <- function() {
  DropTransformation$new(c("col1", "col2"))
}
get_xform_missing <- function() {
  MissingValuesTransformation$new()
}

#' Does an HTML element exist in a shiny page?
#'
#' @param driver A {shinytest2} `AppDriver`
#' @param id An ID
is_shiny_exists <- function(driver, id) {
  driver$get_js(glue::glue("!!document.getElementById('{ id }')"))
}

#' Is an HTML element enabled in a shiny page?
#'
#' @param driver A {shinytest2} `AppDriver`
#' @param id An ID
is_shiny_enabled <- function(driver, id) {
  driver$get_js(glue::glue("!$('#{ id }').is(':disabled')"))
}

#' Is an HTML element visible in a shiny page?
#'
#' @param driver A {shinytest2} `AppDriver`
#' @param id An ID
is_shiny_visible <- function(driver, id) {
  driver$get_js(glue::glue("$('#{ id }').is(':visible')"))
}

#' Wait until an HTML element exists on the page
#'
#' @param driver A {shinytest2} `AppDriver`
#' @param id An ID
wait_for_exists <- function(driver, id) {
  driver$wait_for_js(glue::glue("!!document.getElementById('{ id }')"))
}

#' Wait until an HTML element does not exist on the page
#'
#' @param driver A {shinytest2} `AppDriver`
#' @param id An ID
wait_for_not_exists <- function(driver, id) {
  driver$wait_for_js(glue::glue("!document.getElementById('{ id }')"))
}

#' Wait until an HTML element is visible on the page
#'
#' @param driver A {shinytest2} `AppDriver`
#' @param id An ID
wait_for_visible <- function(driver, id) {
  wait_for_exists(driver, id)
  driver$wait_for_js(glue::glue("$('#{ id }').is(':visible')"))
}

#' Wait until an HTML element is not visible on the page
#'
#' @param driver A {shinytest2} `AppDriver`
#' @param id An ID
wait_for_not_visible <- function(driver, id) {
  wait_for_exists(driver, id)
  driver$wait_for_js(glue::glue("!$('#{ id }').is(':visible')"))
}

#' Wait until an HTML element is enabled on the page
#'
#' @param driver A {shinytest2} `AppDriver`
#' @param id An ID
wait_for_enabled <- function(driver, id) {
  wait_for_exists(driver, id)
  driver$wait_for_js(glue::glue("$('#{ id }').is(':enabled')"))
}

#' Wait until an HTML element is disabled on the page
#'
#' @param driver A {shinytest2} `AppDriver`
#' @param id An ID
wait_for_disabled <- function(driver, id) {
  wait_for_exists(driver, id)
  driver$wait_for_js(glue::glue("$('#{ id }').is(':disabled')"))
}

#' Expect that all export values from a shiny app are NULL
#'
#' @param driver A {shinytest2} `AppDriver`
#' @param which Either `TRUE` to test all export values, or a vector of export values
expect_null_exports <- function(driver, which = TRUE) {
  values <- driver$get_values(export = which)
  testthat::expect_null(unlist(values))
}
