#' Get a default value in case of `NULL`
#'
#' @param x Any value
#' @param y Any value
#' @return The first argument, unless it is `NULL`, in which case the second argument is returned
#' @name nullor
`%||%` <- function (x, y) {
  if (is.null(x)) y else x
}

#' Concatenate with no whitespace
#'
#' @param ... Parameters to pass to `cat()`
cat0 <- function(...) {
  cat(..., sep = "")
}

#' Capitalize the first letter
#'
#' @param s A string
firstup <- function(s) {
  substr(s, 1, 1) <- toupper(substr(s, 1, 1))
  s
}

#' Assign an object to the global environment
#'
#' This is a way to get around CRAN checks that complain of assigning to the globalenv
#' @param name Variable name
#' @param x Value to assign
#' @param pos Do not use this argument - it's meant to always be 1 to assign to globalenv
assign_to_global <- function(name, x, pos = 1L) {
  assign(name, x, envir = as.environment(pos))
}

#' Check if a string is a valid R object name
#'
#' @param s A string
is_valid_name <- function(s) {
  !is.null(s) && make.names(s) == s
}

#' Insert a chunk of code to the currently active RStudio document
#'
#' @param text Code to insert
#' @param newline Whether to insert an additional empty line at the end
#' @param remove_duplicate_libs Whether to remove duplicate `library()` calls
insert_text <- function(text, newline = TRUE, remove_duplicate_libs = TRUE) { # nocov start
  id <- rstudioapi::getSourceEditorContext()$id
  if (is.null(id)) {
    id <- rstudioapi::documentNew("")
  }

  if (newline) {
    if (endsWith(text, "\n\n")) {}
    else if (endsWith(text, "\n")) {
      text <- paste0(text, "\n")
    } else {
      text <- paste0(text, "\n\n")
    }
  }

  tryCatch({

    if (remove_duplicate_libs) {
      existing_contents <- rstudioapi::getSourceEditorContext()$contents
      existing_library_calls <- existing_contents[grepl("^library(.+)$", existing_contents)]
      text <- remove_duplicate_lines(text, existing_library_calls)
    }

    rstudioapi::setCursorPosition(Inf, id = id)
    loc <- rstudioapi::getSourceEditorContext()$selection
    # if the end of the document is not a new line, add a new line
    if (loc[[1]]$range$end[[2]] != 1) {
      text <- paste0("\n", text)
    }
  }, finally = {
    rstudioapi::insertText(id = id, text = text)
  }, error = function(e) {})
  invisible(NULL)
} # nocov end

#' Remove duplicate lines from a piece of text
#'
#' @param text Any text (a single string with line breaks)
#' @param lines_to_remove List of lines to remove. Only complete lines in the input text
#' that match a line to remove will get removed.
remove_duplicate_lines <- function(text = "", lines_to_remove = c()) {
  for (line in lines_to_remove) {
    # Deal with lines in the middle
    text <- gsub(paste0("\n", line, "\n"), "\n", text, fixed = TRUE)

    # Deal with first line
    if (startsWith(text, paste0(line, "\n"))) {
      text <- sub(paste0(line, "\n"), "", text, fixed = TRUE)
    }

    # Deal with last line
    if (endsWith(text, paste0("\n", line))) {
      text <- paste0(text, "\n")
      text <- sub(paste0("\n", line, "\n"), "", text, fixed = TRUE)
    }

    if (text == line) {
      text <- ""
    }
  }
  text
}

#' Whether or not a directory is git writable
#'
#' @param dir A directory
is_git_writable <- function(dir) { # nocov start
  tryCatch({
    owd <- setwd(dir)
    on.exit(setwd(owd), add = TRUE)
    gitcreds::gitcreds_get(use_cache = FALSE, set_cache = FALSE)
    processx::run("git", c("push", "--dry-run", "--force", "--no-verify"))
    TRUE
  }, error = function(err) FALSE)
} # nocov end

#' Normalize a path across all operation systems without any warnings
#'
#' @param path A path
make_path <- function(path) {
  suppressWarnings(normalizePath(path, winslash = "/"))
}

#' Determine if a given path is a subpath of another
#'
#' @param parent The path of the ancestor
#' @param child The path of the child
is_subdir <- function(parent, child) {
  parent <- make_path(parent)
  child <- make_path(child)
  startsWith(child, parent)
}

#' Get the name of a file without its extension
#'
#' @param x A path to a file
get_file_name_no_ext <- function(x) {
  tools::file_path_sans_ext(basename(x))
}

#' Get the name of a dataset if it exists in the global environment
#'
#' @param x The name of a dataframe, either reactive or static
get_data_name_str <- function(x = NULL) {
  tryCatch({
    if (is.null(x)) {
      NULL
    } else if (shiny::is.reactive(x)) {
      x
    } else if (checkmate::test_string(x) && checkmate::test_data_frame(get(x, envir = globalenv()))) {
      x
    } else {
      NULL
    }
  }, error = function(err) {
    NULL
  })
}

#' Run an R expression with no output
#'
#' @param expr The R code to run
quietly <- function(expr) {
  utils::capture.output(suppressWarnings(suppressMessages(suppressPackageStartupMessages(
    expr
  ))))
}
