library(shiny)

skip_on_cran()
skip_on_ci()

test_that("LoadComponentDataSource without data sources support", {
  mockery::stub(is_data_sources_supported, "get_domino_version", "1.0.0", 2)

  app <- shinyApp(
    fluidPage(assistdomino:::LoadComponentDataSource$shiny$ui("test")),
    function(input, output, session) {
      assistdomino:::LoadComponentDataSource$shiny$server("test")
    }
  )
  driver <- shinytest2::AppDriver$new(app, expect_values_screenshot_args = FALSE)

  driver$wait_for_idle()
  expect_true(is_shiny_exists(driver, "test-not-supported"))
  expect_false(is_shiny_exists(driver, "test-data_source"))
})

test_that("LoadComponentDataSource with data sources", {
  mockery::stub(is_data_sources_supported, "get_domino_version", DATASOURCES_MIN_VERSION, 2)

  app <- shinyApp(
    fluidPage(assistdomino:::LoadComponentDataSource$shiny$ui("test")),
    function(input, output, session) {
      Sys.setenv("MOCK_DOMINODATAR" = system.file("mock_datasources", package = "assistdomino"))
      mockery::stub(is_data_sources_supported, "get_domino_version", assistdomino:::DATASOURCES_MIN_VERSION, 2)
      local_mocked_bindings(
        get_data_sources_raw = function(...) jsonlite::fromJSON(system.file("mock_datasources", "data_sources.json", package = "assistdomino"), simplifyDataFrame = FALSE),
        .package = "assistdomino"
      )

      res <- assistdomino:::LoadComponentDataSource$shiny$server("test")
      observe(res$name())
      exportTestValues(
        name = res$name(),
        code = res$code(),
        data = res$data(),
        error = res$error()
      )
    }
  )
  driver <- shinytest2::AppDriver$new(app, expect_values_screenshot_args = FALSE, name = "LoadComponentDataSource")

  driver$wait_for_idle()
  expect_false(is_shiny_exists(driver, "test-not-supported"))
  expect_true(is_shiny_exists(driver, "test-data_source"))
  driver$expect_values(export = TRUE)
  expect_identical(driver$get_text("#test-datasource_ui"), "")

  driver$set_inputs(
    `test-data_source` = "s3",
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  expect_match(driver$get_text("#test-datasource_ui"), "csv")

  driver$set_inputs(
    `test-ds_1-ds_s3-files-file_clicked` = "badfile.csv",
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  driver$expect_values(export = TRUE)

  driver$set_inputs(
    `test-ds_1-ds_s3-files-file_clicked` = "internal/s3file2.csv",
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  driver$expect_values(export = TRUE)

  driver$set_inputs(
    `test-ds_1-ds_s3-files-file_clicked` = "internal/s3file2.csv",
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  driver$expect_values(export = TRUE)

  driver$set_inputs(
    `test-ds_1-ds_s3-files-file_clicked` = "s3file.csv",
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  driver$expect_values(export = TRUE)

  driver$set_inputs(
    `test-data_source` = NULL,
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  expect_identical(driver$get_text("#test-datasource_ui"), "")

  driver$set_inputs(
    `test-data_source` = "gcs",
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  expect_match(driver$get_text("#test-datasource_ui"), "no supported files")

  driver$set_inputs(
    `test-data_source` = "snowflake_empty",
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  expect_true(is_shiny_visible(driver, "test-ds_3-no-warehouse-error"))

  driver$set_inputs(
    `test-data_source` = "snowflake",
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  expect_false(is_shiny_visible(driver, "test-ds_4-no-warehouse-error"))
  expect_identical(driver$get_value(input = "test-ds_4-ds_snowflake-database"), "TESTDB")
  expect_identical(driver$get_value(input = "test-ds_4-ds_snowflake-schema"), "TESTSCHEMA")
  expect_null(driver$get_value(input = "test-ds_4-ds_snowflake-table"))
  driver$set_inputs(
    `test-ds_4-ds_snowflake-table` = "snow1",
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  expect_false(is_shiny_visible(driver, "test-ds_4-ds_snowflake-too-many-rows-error"))
  driver$expect_values(export = TRUE)

  driver$set_inputs(
    `test-ds_4-ds_snowflake-table` = "diamonds",
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  expect_true(is_shiny_visible(driver, "test-ds_4-ds_snowflake-too-many-rows-error"))
  driver$expect_values(export = c("code", "error", "name"))

  driver$set_inputs(
    `test-data_source` = "snowflake_db",
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  expect_false(is_shiny_visible(driver, "test-ds_5-no-warehouse-error"))
  expect_identical(driver$get_value(input = "test-ds_5-ds_snowflake-database"), "TESTDB2")
  expect_null(driver$get_value(input = "test-ds_5-ds_snowflake-schema"))
  expect_null(driver$get_value(input = "test-ds_5-ds_snowflake-table"))
  driver$set_inputs(
    `test-ds_5-ds_snowflake-schema` = "mockschema",
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()

  driver$set_inputs(
    `test-ds_5-ds_snowflake-table` = "snowdb1",
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  driver$expect_values(export = TRUE)

  driver$set_inputs(
    `test-ds_5-ds_snowflake-table` = NULL,
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  driver$expect_values(export = TRUE)

  driver$set_inputs(
    `test-ds_5-ds_snowflake-table` = "snowdbempty",
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  driver$expect_values(export = TRUE)

  driver$set_inputs(
    `test-data_source` = "snowflake_warehouseonly",
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  expect_false(is_shiny_visible(driver, "test-ds_6-no-warehouse-error"))
  expect_null(driver$get_value(input = "test-ds_6-ds_snowflake-database"))
  expect_null(driver$get_value(input = "test-ds_6-ds_snowflake-schema"))
  expect_null(driver$get_value(input = "test-ds_6-ds_snowflake-table"))
  driver$set_inputs(
    `test-ds_6-ds_snowflake-database` = "mockdb",
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  driver$set_inputs(
    `test-ds_6-ds_snowflake-schema` = "mockschema",
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()

  driver$set_inputs(
    `test-ds_6-ds_snowflake-table` = "snowware",
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  driver$expect_values(export = TRUE)
  driver$set_inputs(
    `test-ds_6-ds_snowflake-table` = "snowware2",
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  driver$expect_values(export = TRUE)

  driver$set_inputs(
    `test-data_source` = "redshift",
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  expect_identical(driver$get_value(input = "test-ds_7-ds_redshift-database"), "TESTDB")
  expect_null(driver$get_value(input = "test-ds_7-ds_redshift-schema"))
  expect_null(driver$get_value(input = "test-ds_7-ds_redshift-table"))
  driver$set_inputs(
    `test-ds_7-ds_redshift-schema` = "mockschema",
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  driver$set_inputs(
    `test-ds_7-ds_redshift-table` = "red1",
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  driver$expect_values(export = TRUE)
})
