test_that("ddr_function works for real data sources", {
  withr::local_envvar(c("MOCK_DOMINODATAR" = ""))
  expect_identical(ddr_function("query", text = TRUE), "DominoDataR::query")
  expect_identical(ddr_function("list_keys", text = TRUE), "DominoDataR::list_keys")
  # TODO The last test is commented out because `DominoDataR::list_keys` causes a GitHub Actions failure in R-devel for some reason.
  # Even merely testing for `expect_error(DominoDataR::list_keys, NA)` causes GitHub Actions to fail.
  #expect_identical(ddr_function("list_keys"), DominoDataR::list_keys)
})

test_that("ddr_function works for mock data sources", {
  withr::local_envvar(c("MOCK_DOMINODATAR" = system.file("mock_datasources", package = PACKAGE_NAME)))
  expect_identical(ddr_function("list_keys"), mock_ddr_list_keys, ignore_function_env = TRUE)
  expect_identical(ddr_function("list_keys", text = TRUE), "mock_ddr_list_keys")
})


test_that("mock datasource_client works", {
  withr::local_envvar(c("MOCK_DOMINODATAR" = ""))
  expect_error(mock_ddr_datasource_client(), "path containing mock data")

  withr::local_envvar(c("MOCK_DOMINODATAR" = "badpath"))
  expect_error(mock_ddr_datasource_client(), "path containing mock data")

  withr::local_envvar(c("MOCK_DOMINODATAR" = system.file("mock_datasources", package = PACKAGE_NAME)))
  expect_error(mock_ddr_datasource_client(), NA)
  expect_identical(as.character(mock_ddr_datasource_client()), system.file("mock_datasources", package = PACKAGE_NAME))
  expect_identical(class(mock_ddr_datasource_client()), "MockDataSourceClient")
})

test_that("mock list_keys works", {
  withr::local_envvar(c("MOCK_DOMINODATAR" = system.file("mock_datasources", package = PACKAGE_NAME)))
  client <- mock_ddr_datasource_client()

  expect_error(mock_ddr_list_keys(client), "missing")
  expect_error(mock_ddr_list_keys(client, "nosource"), "folder doesn't exist")
  expect_length(mock_ddr_list_keys(client, "s3"), 3)
  expect_length(mock_ddr_list_keys(client, "gcs"), 1)
})

test_that("mock save_object works", {
  withr::local_envvar(c("MOCK_DOMINODATAR" = system.file("mock_datasources", package = PACKAGE_NAME)))
  client <- mock_ddr_datasource_client()

  withr::with_tempdir({
    expect_error(mock_ddr_save_object(client, "s3", "s3file.csv2", "test.csv"), "object doesn't exist")
    expect_false(file.exists("test.csv"))
    mock_ddr_save_object(client, "s3", "s3file.csv", "test.csv")
    expect_true(file.exists("test.csv"))
    expect_identical(
      unname(tools::md5sum(file.path(client, "s3", "s3file.csv"))),
      unname(tools::md5sum("test.csv"))
    )
    mock_ddr_save_object(client, "s3", "internal/s3file2.csv", "test.csv")
    expect_identical(
      unname(tools::md5sum(file.path(client, "s3", "internal", "s3file2.csv"))),
      unname(tools::md5sum("test.csv"))
    )
  })
})

test_that("mock query works to read tables", {
  withr::local_envvar(c("MOCK_DOMINODATAR" = system.file("mock_datasources", package = PACKAGE_NAME)))
  client <- mock_ddr_datasource_client()

  expect_error(mock_ddr_query(client, "redshift", "select * from snow1"), "table doesn't exist")
  expect_identical(
    mock_ddr_query(client, "snowflake", "select * from snow1"),
    read.csv(file.path(client, "snowflake", "snow1.csv"))
  )
  expect_identical(
    mock_ddr_query(client, "snowflake", "SELEct * FRom snow1"),
    mock_ddr_query(client, "snowflake", "select * from snow1")
  )
  expect_identical(
    mock_ddr_query(client, "snowflake", "select * from snow2"),
    read.csv(file.path(client, "snowflake", "snow2.csv"))
  )
})

test_that("mock query works to get tables info", {
  withr::local_envvar(c("MOCK_DOMINODATAR" = system.file("mock_datasources", package = PACKAGE_NAME)))
  client <- mock_ddr_datasource_client()

  expect_equal(
    mock_ddr_query(client, "redshift", "show tables"),
    data.frame(database_name = "mockdb", schema_name = "mockschema", name = c("red1", "red2"), rows = 2)
  )
})

test_that("get_mock_ddr_folder works", {
  withr::local_envvar(c("MOCK_DOMINODATAR" = system.file("mock_datasources", package = PACKAGE_NAME)))
  client <- mock_ddr_datasource_client()

  expect_error(get_mock_ddr_folder("badclient", "s3"), "mock_ddr_datasource_client")
  expect_error(get_mock_ddr_folder(client, "s5"), "doesn't exist")
  expect_identical(get_mock_ddr_folder(client, "s3"), file.path(client, "s3"))
})

test_that("get_mock_ddr_table_name works", {
  expect_error(get_mock_ddr_table_name("SELECT FROM table1"))
  expect_identical(get_mock_ddr_table_name("SELECT * FROM table1"), "table1")
  expect_identical(get_mock_ddr_table_name("selECT * FRom table2"), "table2")
  expect_identical(get_mock_ddr_table_name("select * from table3 limit 100"), "table3")
  expect_identical(get_mock_ddr_table_name("select * from db.schema.table4"), "table4")
  expect_identical(get_mock_ddr_table_name("select * from db.schema.table5 limit 500"), "table5")
})
