library(shiny)

skip_on_cran()

test_that("page_viz shiny module works when not given a dataset", {
  app <- shinyApp(
    ui = fluidPage(
      assistdomino:::page_viz_ui("test")
    ),
    server = function(input, output, session) {
      Sys.setenv(SHINYMIXPANEL_DISABLE = "1")

      assign("df1", cars, envir = globalenv())

      res <- assistdomino:::page_viz_server("test")
      exportTestValues(
        code = res$code()
      )
    }
  )

  driver <- shinytest2::AppDriver$new(app, expect_values_screenshot_args = FALSE)
  driver$wait_for_idle()
  wait_for_visible(driver, "test-data_select")

  driver$set_inputs(
    `test-data_select_mod-dataframes-file_clicked` = "df1",
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )

  wait_for_visible(driver, "test-main_section")
  driver$wait_for_js('!!document.querySelector(".xyvar")')
  expect_false(is_shiny_enabled(driver, "test-continue"))
  expect_equal(driver$get_value(export = "code"), "")

  skip_on_ci()  # Bug in https://github.com/dreamRs/esquisse causes it to fail in GitHub Actions

  variables <- driver$get_text("#test-plot_module-dragvars-source")
  variables <- trimws(variables)
  variables <- gsub("\\s+", " ", variables)
  expect_equal(variables, "speed dist")

  xbox_id <- driver$get_js('document.querySelector(".xyvar").id')
  xbox_id <- sub(".*-(.*)", "\\1", xbox_id)
  driver$set_inputs(
    `test-plot_module-dragvars:esquisse.dragula` = list(
      target = as.list(setNames(
        "dist", xbox_id
      ))
    ),
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  wait_for_enabled(driver, "test-continue")

  code <- driver$get_value(export = "code")
  expect_match(code, "library\\(ggplot2\\).*plot <- ggplot\\(df1\\).*x = dist")
})

test_that("page_viz shiny module works when given a static dataset", {
  app <- shinyApp(
    ui = fluidPage(
      assistdomino:::page_viz_ui("test")
    ),
    server = function(input, output, session) {
      Sys.setenv(SHINYMIXPANEL_DISABLE = "1")

      assign("df1", cars, envir = globalenv())

      res <- assistdomino:::page_viz_server("test", data_name_in = "df1")
      exportTestValues(
        code = res$code()
      )
    }
  )

  driver <- shinytest2::AppDriver$new(app, expect_values_screenshot_args = FALSE)
  driver$wait_for_idle()

  wait_for_visible(driver, "test-main_section")
  expect_false(is_shiny_visible(driver, "test-data_select"))
  driver$wait_for_js('!!document.querySelector(".xyvar")')
  expect_false(is_shiny_enabled(driver, "test-continue"))
  expect_equal(driver$get_value(export = "code"), "")

  skip_on_ci()

  variables <- driver$get_text("#test-plot_module-dragvars-source")
  variables <- trimws(variables)
  variables <- gsub("\\s+", " ", variables)
  expect_equal(variables, "speed dist")

  xbox_id <- driver$get_js('document.querySelector(".xyvar").id')
  xbox_id <- sub(".*-(.*)", "\\1", xbox_id)
  driver$set_inputs(
    `test-plot_module-dragvars:esquisse.dragula` = list(
      target = as.list(setNames(
        "speed", xbox_id
      ))
    ),
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  wait_for_enabled(driver, "test-continue")

  code <- driver$get_value(export = "code")
  expect_match(code, "library\\(ggplot2\\).*plot <- ggplot\\(df1\\).*x = speed")
})

test_that("page_viz shiny module works when given a reactive dataset", {
  app <- shinyApp(
    ui = fluidPage(
      selectInput("data", "data", c("df1", "df2")),
      assistdomino:::page_viz_ui("test")
    ),
    server = function(input, output, session) {
      Sys.setenv(SHINYMIXPANEL_DISABLE = "1")

      assign("df1", cars, envir = globalenv())
      assign("df2", iris, envir = globalenv())

      res <- assistdomino:::page_viz_server("test", data_name_in = reactive(input$data))
      exportTestValues(
        code = res$code()
      )
    }
  )

  driver <- shinytest2::AppDriver$new(app, expect_values_screenshot_args = FALSE)
  driver$wait_for_idle()

  wait_for_visible(driver, "test-main_section")
  expect_false(is_shiny_visible(driver, "test-data_select"))
  driver$wait_for_js('!!document.querySelector(".xyvar")')
  expect_false(is_shiny_enabled(driver, "test-continue"))
  expect_equal(driver$get_value(export = "code"), "")

  skip_on_ci()

  variables <- driver$get_text("#test-plot_module-dragvars-source")
  variables <- trimws(variables)
  variables <- gsub("\\s+", " ", variables)
  expect_equal(variables, "speed dist")

  xbox_id <- driver$get_js('document.querySelector(".xyvar").id')
  xbox_id <- sub(".*-(.*)", "\\1", xbox_id)
  driver$set_inputs(
    `test-plot_module-dragvars:esquisse.dragula` = list(
      target = as.list(setNames(
        "dist", xbox_id
      ))
    ),
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  wait_for_enabled(driver, "test-continue")

  expect_match(
    driver$get_value(export = "code"),
    "library\\(ggplot2\\).*plot <- ggplot\\(df1\\).*x = dist"
  )

  driver$set_inputs("test-plot_name" = "myplot")
  driver$wait_for_idle()
  expect_match(
    driver$get_value(export = "code"),
    "library\\(ggplot2\\).*myplot <- ggplot\\(df1\\).*x = dist"
  )

  driver$set_inputs("data" = "df2")
  driver$wait_for_js('$(".label-esquisse").first().text() == "Sepal.Length"')

  xbox_id <- driver$get_js('document.querySelector(".xyvar").id')
  xbox_id <- sub(".*-(.*)", "\\1", xbox_id)
  driver$set_inputs(
    `test-plot_module-dragvars:esquisse.dragula` = list(
      target = as.list(setNames(
        "Sepal.Width", xbox_id
      ))
    ),
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  expect_match(
    driver$get_value(export = "code"),
    "library\\(ggplot2\\).*myplot <- ggplot\\(df2\\).*x = Sepal.Width"
  )

  driver$set_inputs("test-plot_name" = "bad name")
  driver$wait_for_idle()
  code <- driver$wait_for_value(export = "code")
  expect_match(
    code,
    "library\\(ggplot2\\).*bad.name <- ggplot\\(df2\\).*x = Sepal.Width"
  )
})
