library(shiny)

skip_on_cran()

test_that("transformation-modal works", {
  app <- shinyApp(
    ui <- fluidPage(
      selectInput("action", "action", c("add", "edit", "insert")),
      selectInput("xform", "xform", c("none", "select", "drop", "missing", "filter", "agg", "merge")),
      actionButton("show_modal", "show modal")
    ),
    server = function(input, output, session) {
      cars2 <- mtcars
      names(cars2) <- paste0(names(mtcars), "2")
      names(cars2)[1:2] <- c("mpg", "cyl")
      assign("cars2", cars2, envir = globalenv())

      modal <- assistdomino:::transformation_modal("test")

      xform <- reactive({
        if (input$xform == "none") {
          NULL
        } else if (input$xform == "select") {
          assistdomino:::SelectTransformation$new(c("mpg", "vs"), "seldf")
        } else if (input$xform == "drop") {
          assistdomino:::DropTransformation$new(c("cyl", "hp"), "dropdf")
        } else if (input$xform == "missing") {
          assistdomino:::MissingValuesTransformation$new(name_out = "misdf")
        } else if (input$xform == "filter") {
          assistdomino:::FilterTransformation$new("cyl", ">=", 6, name_out = "filtdf")
        } else if (input$xform == "agg") {
          assistdomino:::AggregateTransformation$new("cyl", c("hp" = "mean", "drat" = "max"), name_out = "aggdf")
        } else if (input$xform == "merge") {
          assistdomino:::MergeTransformation$new("cars2", type = "outer", by = c("vs" = "vs2", "cyl" = "cyl"), name_out = "mergedf")
        }
      })

      meta <- reactive({
        if (input$action == "add") {
          1
        } else if (input$action == "edit") {
          2
        } else if (input$action == "insert") {
          3
        }
      })

      observeEvent(input$show_modal, {
        modal$show(mtcars, input$action, xform = xform(), meta = meta())
      })

      exportTestValues(
        xform = { if (is.null(modal$result())) NULL else utils::capture.output(modal$result()) },
        action = modal$action(),
        meta = modal$meta()
      )
    }
  )

  driver <- shinytest2::AppDriver$new(app, expect_values_screenshot_args = FALSE, name = "mod-transformation-modal")
  driver$wait_for_idle()

  driver$click("show_modal")
  wait_for_visible(driver, "test-xform_type")
  expect_identical(driver$get_value(input = "test-xform_type"), "select")
  wait_for_disabled(driver, "test-apply")
  driver$set_inputs("test-select-select_cols" = c("vs", "am"))
  wait_for_enabled(driver, "test-apply")
  driver$click("test-apply")
  wait_for_not_exists(driver, "test-apply")
  driver$expect_values(export = TRUE)

  driver$set_inputs("xform" = "drop")
  driver$wait_for_idle()
  driver$click("show_modal")
  wait_for_visible(driver, "test-apply")
  expect_false(is_shiny_visible(driver, "test-xform_type"))
  driver$click("test-apply")
  wait_for_not_exists(driver, "test-apply")
  driver$expect_values(export = TRUE)

  driver$set_inputs("xform" = "none")
  driver$wait_for_idle()
  driver$click("show_modal")
  wait_for_visible(driver, "test-apply")
  expect_true(is_shiny_visible(driver, "test-xform_type"))
  driver$set_inputs("test-xform_type" = "missing")
  driver$wait_for_idle()
  driver$click("test-apply")
  wait_for_not_exists(driver, "test-apply")
  driver$expect_values(export = TRUE)

  driver$set_inputs("xform" = "select")
  driver$wait_for_idle()
  driver$click("show_modal")
  wait_for_visible(driver, "test-apply")
  expect_identical(driver$get_value(input = "test-select-select_cols"), c("mpg", "vs"))
  expect_identical(driver$get_value(input = "test-select-select_name"), "seldf")
  wait_for_enabled(driver, "test-apply")
  driver$set_inputs("test-select-select_cols" = "")
  wait_for_disabled(driver, "test-apply")
  driver$set_inputs("test-select-select_cols" = c("gear", "carb"))
  wait_for_enabled(driver, "test-apply")
  driver$set_inputs("test-select-select_name" = "one two")
  driver$wait_for_idle()
  expect_identical(driver$get_value(input = "test-select-select_name"), "one.two")
  driver$expect_values(export = TRUE)
  driver$click("test-apply")
  wait_for_not_exists(driver, "test-apply")
  driver$expect_values(export = TRUE)

  driver$set_inputs("xform" = "filter")
  driver$wait_for_idle()
  driver$click("show_modal")
  wait_for_visible(driver, "test-apply")
  expect_identical(driver$get_value(input = "test-filter-filter_col"), "cyl")
  expect_identical(driver$get_value(input = "test-filter-filter_op"), ">=")
  expect_identical(driver$get_value(input = "test-filter-filter_value"), "6")
  wait_for_enabled(driver, "test-apply")
  driver$set_inputs(
    "test-filter-filter_col" = "disp",
    "test-filter-filter_op" = "<",
    "test-filter-filter_value" = "100"
  )
  wait_for_enabled(driver, "test-apply")
  driver$set_inputs("test-filter-filter_name" = "smalls!")
  driver$wait_for_idle()
  expect_identical(driver$get_value(input = "test-filter-filter_name"), "smalls.")
  driver$click("test-apply")
  wait_for_not_exists(driver, "test-apply")
  driver$expect_values(export = TRUE)

  driver$set_inputs("xform" = "missing", "action" = "insert")
  driver$wait_for_idle()
  driver$click("show_modal")
  wait_for_visible(driver, "test-apply")
  expect_identical(driver$get_value(input = "test-missing-missing_name"), "misdf")
  expect_identical(driver$get_value(input = "test-missing-missing_type"), "all")
  driver$set_inputs("test-missing-missing_type" = "column", "test-missing-missing_col" = "drat")
  driver$wait_for_idle()
  driver$click("test-apply")
  wait_for_not_exists(driver, "test-apply")
  driver$expect_values(export = TRUE)

  driver$set_inputs("action" = "edit", "xform" = "agg")
  driver$wait_for_idle()
  driver$click("show_modal")
  wait_for_visible(driver, "test-apply")
  expect_identical(driver$get_value(input = "test-aggregate-aggregate_name"), "aggdf")
  expect_identical(driver$get_value(input = "test-aggregate-aggregate_cols"), "cyl")
  driver$set_inputs("test-aggregate-aggregate_cols" = c("disp", "carb"))
  expect_identical(
    jsonlite::fromJSON(driver$get_value(input = "test-aggregate-aggregate_pairs-existing_pairs_raw"), simplifyDataFrame = FALSE),
    list(list(hp = "mean"), list(drat = "max"))
  )
  driver$set_inputs(
    `test-aggregate-aggregate_pairs-remove_pair` = 2,
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  expect_identical(
    jsonlite::fromJSON(driver$get_value(input = "test-aggregate-aggregate_pairs-existing_pairs_raw"), simplifyDataFrame = FALSE),
    list(list(hp = "mean"))
  )
  driver$set_inputs(
    `test-aggregate-aggregate_pairs-remove_pair` = 1,
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  expect_identical(
    jsonlite::fromJSON(driver$get_value(input = "test-aggregate-aggregate_pairs-existing_pairs_raw"), simplifyDataFrame = FALSE),
    list()
  )
  driver$set_inputs("test-aggregate-aggregate_pairs-choice_1" = "hp", "test-aggregate-aggregate_pairs-choice_2" = "mean")
  driver$wait_for_idle()
  expect_identical(
    jsonlite::fromJSON(driver$get_value(input = "test-aggregate-aggregate_pairs-existing_pairs_raw"), simplifyDataFrame = FALSE),
    list(list(hp = "mean"))
  )
  driver$set_inputs("test-aggregate-aggregate_pairs-choice_1" = "drat", "test-aggregate-aggregate_pairs-choice_2" = "max")
  driver$wait_for_idle()
  expect_identical(
    jsonlite::fromJSON(driver$get_value(input = "test-aggregate-aggregate_pairs-existing_pairs_raw"), simplifyDataFrame = FALSE),
    list(list(hp = "mean"), list(drat = "max"))
  )
  driver$set_inputs("test-aggregate-aggregate_pairs-choice_1" = "drat", "test-aggregate-aggregate_pairs-choice_2" = "mean")
  driver$wait_for_idle()
  expect_identical(
    jsonlite::fromJSON(driver$get_value(input = "test-aggregate-aggregate_pairs-existing_pairs_raw"), simplifyDataFrame = FALSE),
    list(list(hp = "mean"), list(drat = "max"), list(drat = "mean"))
  )
  driver$set_inputs("test-aggregate-aggregate_pairs-choice_1" = "hp", "test-aggregate-aggregate_pairs-choice_2" = "mean")
  driver$wait_for_idle()
  expect_identical(
    jsonlite::fromJSON(driver$get_value(input = "test-aggregate-aggregate_pairs-existing_pairs_raw"), simplifyDataFrame = FALSE),
    list(list(hp = "mean"), list(drat = "max"), list(drat = "mean"))
  )
  driver$click("test-apply")
  wait_for_not_exists(driver, "test-apply")
  driver$expect_values(export = TRUE)

  driver$set_inputs("action" = "add", "xform" = "merge")
  driver$wait_for_idle()
  driver$click("show_modal")
  wait_for_visible(driver, "test-apply")
  expect_identical(driver$get_value(input = "test-merge-name"), "mergedf")
  expect_identical(driver$get_value(input = "test-merge-df"), "cars2")
  expect_identical(
    jsonlite::fromJSON(driver$get_value(input = "test-merge-merge_pairs-existing_pairs_raw"), simplifyDataFrame = FALSE),
    list(list(vs = "vs2"), list(cyl = "cyl"))
  )
  driver$set_inputs("test-merge-merge_pairs-choice_1" = "am", "test-merge-merge_pairs-choice_2" = "am2")
  driver$set_inputs(
    `test-merge-merge_pairs-remove_pair` = 1,
    allow_no_input_binding_ = TRUE, priority_ = "event"
  )
  driver$wait_for_idle()
  driver$click("test-apply")
  wait_for_not_exists(driver, "test-apply")
  driver$expect_values(export = TRUE)

  driver$set_inputs("action" = "edit", "xform" = "merge")
  driver$wait_for_idle()
  driver$click("show_modal")
  wait_for_visible(driver, "test-apply")
  driver$set_inputs(
    "test-merge-type" = "right",
    "test-merge-natural" = TRUE
  )
  driver$click("test-apply")
  wait_for_not_exists(driver, "test-apply")
  driver$expect_values(export = TRUE)
})
