test_that("get_user_project_dir works", {
  expect_identical(
    withr::with_envvar(
      c("DOMINO_WORKING_DIR" = "a/b/c"),
      get_user_project_dir()
    ),
    "a/b/c"
  )

  expect_identical(
    withr::with_envvar(
      c("DOMINO_WORKING_DIR" = NA),
      get_user_project_dir()
    ),
    getwd()
  )
})

test_that("get_user_datasets_dir works", {
  expect_identical(
    withr::with_envvar(
      c("DOMINO_DATASETS_DIR" = "a/b/c"),
      get_user_datasets_dir()
    ),
    "a/b/c"
  )

  expect_identical(
    withr::with_envvar(
      c("DOMINO_DATASETS_DIR" = NA),
      get_user_datasets_dir()
    ),
    getwd()
  )
})

test_that("get_user_git_dir works", {
  expect_identical(
    withr::with_envvar(
      c("DOMINO_IMPORTED_CODE_DIR" = "a/b/c"),
      get_user_git_dir()
    ),
    "a/b/c"
  )

  expect_identical(
    withr::with_envvar(
      c("DOMINO_IMPORTED_CODE_DIR" = NA),
      get_user_git_dir()
    ),
    ""
  )
})

test_that("get_user_git_repos works", {
  mockery::stub(get_user_git_repos, "get_user_git_dir", "")
  expect_identical(get_user_git_repos(), character(0))

  mockery::stub(get_user_git_repos, "get_user_git_dir",
                system.file("tests_data", "git_repos", package = PACKAGE_NAME))
  expect_identical(basename(get_user_git_repos()), c("repo1", "repo2", "repo3"))
})

test_that("get_imported_projects_dirs", {
  # remove all existing project dir envvars
  projects <- grep("^DOMINO_.*_WORKING_DIR$", names(Sys.getenv()), value = TRUE)
  for(proj in projects) {
    withr::local_envvar(setNames(NA, proj))
  }

  expect_identical(get_imported_projects_dirs(), character(0))

  withr::local_envvar(c("DOMINO_PROJ1_WORKING_DIR" = "a/b/c"))
  expect_identical(get_imported_projects_dirs(), c("a/b/c"))

  withr::local_envvar(c("DOMINO_PROJ2_WORKING_DIR" = "a/b/d"))
  expect_identical(get_imported_projects_dirs(), c("a/b/c", "a/b/d"))
})

test_that("get_project_name works", {
  expect_identical(
    withr::with_envvar(
      c("DOMINO_PROJECT_NAME" = "my great project"),
      get_project_name()
    ),
    "my great project"
  )

  expect_identical(
    withr::with_envvar(
      c("DOMINO_PROJECT_NAME" = NA),
      get_project_name()
    ),
    "sample_project"
  )
})

test_that("get_user_upload_dir works", {
  mockery::stub(get_user_upload_dir, "get_user_datasets_dir", "/path/to/datasets")
  mockery::stub(get_user_upload_dir, "get_project_name", "test_proj")

  expect_identical(
    withr::with_envvar(
      c("DOMINO_IS_GIT_BASED" = NA),
      get_user_upload_dir()
    ),
    file.path("/path/to/datasets", "local", "test_proj")
  )

  expect_identical(
    withr::with_envvar(
      c("DOMINO_IS_GIT_BASED" = "true"),
      get_user_upload_dir()
    ),
    file.path("/path/to/datasets", "test_proj")
  )
})

test_that("get_api_key works", {
  expect_error(
    withr::with_envvar(
      c("DOMINO_USER_API_KEY" = NA),
      get_api_key()
    )
  )

  expect_identical(
    withr::with_envvar(
      c("DOMINO_USER_API_KEY" = "a1b2c3d4"),
      get_api_key()
    ),
    "a1b2c3d4"
  )
})

test_that("get_api_base works", {
  expect_error(
    withr::with_envvar(
      c("DOMINO_API_HOST" = NA),
      get_api_base()
    )
  )

  expect_identical(
    withr::with_envvar(
      c("DOMINO_API_HOST" = "https://test.dominodatalab.com"),
      get_api_base()
    ),
    "https://test.dominodatalab.com"
  )
})

test_that("get_domino_version works", {
  httptest::with_mock_api({
    withr::local_envvar(c("DOMINO_USER_API_KEY" = "testkey",
                          "DOMINO_API_HOST" = "https://trial.dominodatalab.com"))

    mockery::stub(get_domino_version, "call_api", function(...) stop("test"))
    expect_identical(get_domino_version(), "0")
    mockery::stub(get_domino_version, "call_api", call_api)

    expect_identical(get_domino_version(), "1.2.3")
  })

  expect_identical(httptest::without_internet(get_domino_version()), "0")
})

test_that("get_user_id works", {
  httptest::with_mock_api({
    withr::local_envvar(c("DOMINO_USER_API_KEY" = "testkey",
                          "DOMINO_API_HOST" = "https://trial.dominodatalab.com"))

    mockery::stub(get_user_id, "call_api", function(...) stop("test"))
    expect_identical(get_user_id(), "0")
    mockery::stub(get_user_id, "call_api", call_api)

    expect_identical(get_user_id(), "123456abcdef")
  })

  expect_identical(httptest::without_internet(get_user_id()), "0")
})

test_that("get_mixpanel_enabled settings work", {
  httptest::with_mock_api({
    withr::local_envvar(c("DOMINO_USER_API_KEY" = "testkey",
                          "DOMINO_API_HOST" = "https://trial.dominodatalab.com"))

    mockery::stub(get_mixpanel_enabled_client, "call_api", function(...) stop("test"))
    mockery::stub(get_mixpanel_enabled_server, "call_api", function(...) stop("test"))
    expect_false(get_mixpanel_enabled_client())
    expect_false(get_mixpanel_enabled_server())
    mockery::stub(get_mixpanel_enabled_client, "call_api", call_api)
    mockery::stub(get_mixpanel_enabled_server, "call_api", call_api)

    expect_true(get_mixpanel_enabled_client())
    expect_true(get_mixpanel_enabled_server())
  })

  expect_false(httptest::without_internet(get_mixpanel_enabled_client()))
  expect_false(httptest::without_internet(get_mixpanel_enabled_server()))
})

test_that("is_data_sources_supported works", {
  mockery::stub(is_data_sources_supported, "get_domino_version", "5.4")
  expect_false(is_data_sources_supported())
  mockery::stub(is_data_sources_supported, "get_domino_version", "5.5")
  expect_true(is_data_sources_supported())
  mockery::stub(is_data_sources_supported, "get_domino_version", "5.6")
  expect_true(is_data_sources_supported())
})

test_that("get_data_sources_raw works", {
  httptest::with_mock_api({
    withr::local_envvar(c("DOMINO_USER_API_KEY" = "testkey",
                          "DOMINO_API_HOST" = "https://trial.dominodatalab.com"))

    expect_length(
      withr::with_envvar(
        c("DOMINO_PROJECT_ID" = "projbad"),
        get_data_sources_raw()
      ),
      0
    )

    expect_length(
      withr::with_envvar(
        c("DOMINO_PROJECT_ID" = "proj1"),
        get_data_sources_raw()
      ),
      3
    )

    expect_identical(
      withr::with_envvar(
        c("DOMINO_PROJECT_ID" = "proj1"),
        get_data_sources_raw()[[1]]$id
      ),
      "123abc"
    )

    expect_identical(
      withr::with_envvar(
        c("DOMINO_PROJECT_ID" = "proj1"),
        get_data_sources_raw()[[1]]$name
      ),
      "s3"
    )

    expect_identical(
      withr::with_envvar(
        c("DOMINO_PROJECT_ID" = "proj1"),
        get_data_sources_raw()[[2]]$name
      ),
      "snowflake"
    )

    expect_length(
      withr::with_envvar(
        c("DOMINO_PROJECT_ID" = "mockproj"),
        get_data_sources_raw()
      ),
      7
    )
  })
})

test_that("get_data_sources works", {
  mockery::stub(get_data_sources, "get_data_sources_raw", list())
  expect_length(get_data_sources(), 0)

  mockery::stub(get_data_sources, "get_data_sources_raw", list(list(id = "id", name = "name", dataSourceType = "S3Config", config = list())))
  expect_length(get_data_sources(), 1)

  mockery::stub(get_data_sources, "get_data_sources_raw", list(list(id = "id", name = "name", type = "S3Config", config = list())))
  expect_length(get_data_sources(), 0)

  mockery::stub(get_data_sources, "get_data_sources_raw", jsonlite::fromJSON(system.file("mock_datasources", "data_sources.json", package = "assistdomino"), simplifyDataFrame = FALSE))
  sources <- get_data_sources()
  expect_length(sources, 7)
  expect_s3_class(sources[[1]], "DominoDataSource")
  expect_identical(sources[[1]]$id, "123abc")
  expect_identical(sources[[1]]$type, "S3")
  expect_identical(sources[[1]]$name, "s3")
  expect_s3_class(sources[[2]], "DominoDataSource")
  expect_identical(sources[[2]]$id, "123456")
  expect_identical(sources[[2]]$type, "GCS")
  expect_identical(sources[[2]]$name, "gcs")
})
